@@ -469,7 +469,8 @@ protect_require_yca(struct got_object_id *tip_id,
 			break;
 		}
 
-		if (got_object_idset_num_elements(traversed_set) >=
+		if (max_commits_to_traverse > 0 &&
+		    got_object_idset_num_elements(traversed_set) >=
 		    max_commits_to_traverse)
 			break;
 
@@ -841,6 +842,7 @@ copy_ref_delta(int infd, int outfd, off_t *outsize, BUF *buf, size_t *buf_pos,
 		return err;
 
 	*buf_pos += SHA1_DIGEST_LENGTH;
+	*outsize += SHA1_DIGEST_LENGTH;
 	return NULL;
 }
 
@@ -1539,17 +1541,26 @@ protect_refs_from_moving(void)
 
 		RB_FOREACH(pe, got_pathlist_head,
 		    &repo_write.protected_branch_namespaces) {
-			err = protect_ref_namespace(refname, pe->path);
+			if (strcmp(pe->path, refname) != 0)
+				continue;
+
+			err = protect_require_yca(&ref_update->new_id, 0,
+			    NULL, NULL, ref_update->ref);
 			if (err)
 				return err;
+			break;
 		}
 
 		RB_FOREACH(pe, got_pathlist_head,
 		    &repo_write.protected_branches) {
-			if (strcmp(refname, pe->path) == 0) {
-				return got_error_fmt(GOT_ERR_REF_PROTECTED,
-				    "%s", refname);
-			}
+			if (strcmp(pe->path, refname) != 0)
+				continue;
+
+			err = protect_require_yca(&ref_update->new_id, 0,
+			    NULL, NULL, ref_update->ref);
+			if (err)
+				return err;
+			break;
 		}
 	}
 
@@ -1930,17 +1941,53 @@ print_commits(struct got_object_id *root_id, struct got_object_id *end_id,
     struct got_repository *repo, int fd)
 {
 	const struct got_error *err;
-	struct got_commit_graph *graph;
+	struct got_commit_graph *graph = NULL;
 	struct got_object_id_queue reversed_commits;
 	struct got_object_qid *qid;
 	struct got_commit_object *commit = NULL;
 	struct got_pathlist_head changed_paths;
 	int ncommits = 0;
 	const int shortlog_threshold = 50;
+	struct got_object_id *yca_id = NULL;
+	struct got_reference *head_ref = NULL;
+	struct got_object_id *head_id = NULL;
 
 	STAILQ_INIT(&reversed_commits);
 	RB_INIT(&changed_paths);
 
+	if (end_id && got_object_id_cmp(root_id, end_id) != 0) {
+		err = got_commit_graph_find_youngest_common_ancestor(
+		    &yca_id, root_id, end_id, 1, 0, repo_write.repo,
+		    check_cancelled, NULL);
+		if (err)
+			return err;
+		if (yca_id)
+			end_id = yca_id;
+	}
+
+	if (end_id == NULL) {
+		err = got_ref_open(&head_ref, repo_write.repo, GOT_REF_HEAD, 0);
+		if (err) {
+			if (err->code != GOT_ERR_NOT_REF)
+				goto done;
+		} else {
+			err = got_ref_resolve(&head_id, repo_write.repo,
+			    head_ref);
+			if (err)
+				goto done;
+			if (got_object_id_cmp(root_id, head_id) != 0) {
+				err =
+				got_commit_graph_find_youngest_common_ancestor(
+				    &yca_id, root_id, head_id, 1, 0,
+				    repo_write.repo, check_cancelled, NULL);
+				if (err)
+					goto done;
+				if (yca_id)
+					end_id = yca_id;
+			}
+		}
+	}
+
 	/* XXX first-parent only for now */
 	err = got_commit_graph_open(&graph, "/", 1);
 	if (err)
@@ -2005,11 +2052,16 @@ print_commits(struct got_object_id *root_id, struct got_object_id *end_id,
 		got_pathlist_free(&changed_paths, GOT_PATHLIST_FREE_ALL);
 	}
 done:
+	free(yca_id);
+	free(head_id);
+	if (head_ref)
+		got_ref_close(head_ref);
 	if (commit)
 		got_object_commit_close(commit);
 	got_object_id_queue_free(&reversed_commits);
 	got_pathlist_free(&changed_paths, GOT_PATHLIST_FREE_ALL);
-	got_commit_graph_close(graph);
+	if (graph)
+		got_commit_graph_close(graph);
 	return err;
 }
 
@@ -2122,7 +2174,7 @@ static const struct got_error *
 notify_created_ref(const char *refname, struct got_object_id *id,
     struct gotd_imsgev *iev, int fd)
 {
-	const struct got_error *err;
+	const struct got_error *err = NULL;
 	int obj_type;
 
 	err = got_object_get_type(&obj_type, repo_write.repo, id);
