@@ -91,6 +91,7 @@ static struct gotd_session_write {
 	size_t num_notification_refs_received;
 	struct got_pathlist_head *notification_refs_cur;
 	struct gotd_notification_targets notification_targets;
+	int repo_child_packfd;
 } gotd_session;
 
 static struct gotd_session_client {
@@ -222,6 +223,7 @@ recv_packfile_received(int *pack_empty, struct imsg *imsg)
 static const struct got_error *
 request_gotsys_conf(struct gotd_imsgev *iev)
 {
+	const struct got_error *err = NULL;
 	struct gotd_imsg_packfile_get_content content_req;
 	const char *refname = "refs/heads/main";
 	const char *path = "gotsys.conf";
@@ -229,8 +231,9 @@ request_gotsys_conf(struct gotd_imsgev *iev)
 	size_t len;
 	int fd = -1;
 
-	if (ftruncate(gotd_session.content_fd, 0L) == -1)
-		return got_error_from_errno("ftruncate");
+	err = got_opentemp_truncatefd(gotd_session.content_fd);
+	if (err)
+		return err;
 	
 	len = sizeof(content_req) + strlen(refname) + strlen(path);
 	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_PACKFILE_GET_CONTENT,
@@ -661,8 +664,6 @@ forward_notification(struct gotd_session_client *client, struct imsg *imsg)
 	if (notif == NULL)
 		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	STAILQ_REMOVE(&notifications, notif, gotd_session_notif, entry);
-
 	if (notif->action != icontent.action || notif->fd == -1 ||
 	    strcmp(notif->refname, refname) != 0) {
 		err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -745,9 +746,6 @@ forward_notification(struct gotd_session_client *client, struct imsg *imsg)
 	imsg_close(&iev->ibuf, wbuf);
 	gotd_imsg_event_add(iev);
 done:
-	if (notif->fd != -1)
-		close(notif->fd);
-	free(notif);
 	free(refname);
 	free(id_str);
 	return err;
@@ -983,6 +981,33 @@ recv_notification_content(struct imsg *imsg)
 	return NULL;
 }
 
+static const struct got_error *
+report_ref_updates(int do_notify)
+{
+	const struct got_error *err;
+	struct gotd_session_notif *notif;
+
+	err = send_refs_updated(&gotd_session.parent_iev);
+	if (err) {
+		log_warn("%s", err->msg);
+		return err;
+	}
+
+	if (do_notify) {
+		notif = STAILQ_FIRST(&notifications);
+		if (notif == NULL)
+			return NULL;
+
+		err = request_notification(notif);
+		if (err) {
+			log_warn("could not send notification: %s", err->msg);
+			return err;
+		}
+	}
+
+	return NULL;
+}
+
 static void
 session_dispatch_repo_child(int fd, short event, void *arg)
 {
@@ -1085,8 +1110,6 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 			else
 				disconnect(client);
 		} else {
-			struct gotd_session_notif *notif;
-
 			if (do_packfile_verification) {
 				err = verify_packfile(iev);
 			} else if (do_content_verification) {
@@ -1109,27 +1132,17 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 			if (do_ref_update && client->nref_updates > 0) {
 				client->nref_updates--;
 				if (client->nref_updates == 0) {
-					err = send_refs_updated(
-					    &gotd_session.parent_iev);
-					if (err) {
-						log_warn("%s", err->msg);
+					err = report_ref_updates(do_notify);
+					if (err)
 						shut = 1;
-					}
-				}
-			}
-
-			notif = STAILQ_FIRST(&notifications);
-			if (notif && do_notify) {
-				/* Request content for next notification. */
-				err = request_notification(notif);
-				if (err) {
-					log_warn("could not send notification: "
-					    "%s", err->msg);
-					shut = 1;
 				}
 			}
 		}
+
 		imsg_free(&imsg);
+
+		if (err)
+			break;
 	}
 done:
 	if (!shut) {
@@ -1263,9 +1276,6 @@ static const struct got_error *
 recv_packfile(struct gotd_session_client *client)
 {
 	const struct got_error *err = NULL;
-	struct gotd_imsg_recv_packfile ipack;
-	char *basepath = NULL, *pack_path = NULL, *idx_path = NULL;
-	int packfd = -1, idxfd = -1;
 	int pipe[2] = { -1, -1 };
 
 	if (client->packfile_path) {
@@ -1276,23 +1286,49 @@ recv_packfile(struct gotd_session_client *client)
 	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe) == -1)
 		return got_error_from_errno("socketpair");
 
-	/* Send pack pipe end 0 to repo child process. */
-	if (gotd_imsg_compose_event(&gotd_session.repo_child_iev,
-	    GOTD_IMSG_PACKFILE_PIPE, GOTD_PROC_SESSION_WRITE, pipe[0],
-	        NULL, 0) == -1) {
+	/*
+	 * Send pack data pipe end 0 to gotsh(1) (expects just an fd, no data).
+	 *
+	 * We will forward the other pipe end to the repo_write process only
+	 * once we have confirmation that gotsh(1) has received its end.
+	 */
+	if (gotd_imsg_compose_event(&client->iev,
+	    GOTD_IMSG_PACKFILE_PIPE, GOTD_PROC_GOTD, pipe[0], NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose PACKFILE_PIPE");
 		goto done;
 	}
 	pipe[0] = -1;
 
-	/* Send pack pipe end 1 to gotsh(1) (expects just an fd, no data). */
-	if (gotd_imsg_compose_event(&client->iev,
-	    GOTD_IMSG_PACKFILE_PIPE, GOTD_PROC_SESSION_WRITE, pipe[1],
-	    NULL, 0) == -1) {
-		err = got_error_from_errno("imsg compose PACKFILE_PIPE");
-		goto done;
-	}
+	gotd_session.repo_child_packfd = pipe[1];
 	pipe[1] = -1;
+done:
+	if (pipe[0] != -1 && close(pipe[0]) == -1 && err == NULL)
+		err = got_error_from_errno("close");
+	if (pipe[1] != -1 && close(pipe[1]) == -1 && err == NULL)
+		err = got_error_from_errno("close");
+	return err;
+}
+
+static const struct got_error *
+send_packfds_to_repo_child(void)
+{
+	const struct got_error *err = NULL;
+	struct gotd_session_client *client = &gotd_session_client;
+	char *basepath = NULL, *pack_path = NULL, *idx_path = NULL;
+	int packfd = -1, idxfd = -1;
+	struct gotd_imsg_recv_packfile ipack;
+
+	/*
+	 * gotsh(1) has received its end of the pack pipe.
+	 * Send pack pipe end 1 to repo child process.
+	 */
+	if (gotd_imsg_compose_event(
+	    &gotd_session.repo_child_iev,
+	    GOTD_IMSG_PACKFILE_PIPE, GOTD_PROC_SESSION_WRITE,
+	    gotd_session.repo_child_packfd, NULL, 0) == -1)
+		return got_error_from_errno("imsg compose PACKFILE_PIPE");
+
+	gotd_session.repo_child_packfd = -1;
 
 	if (asprintf(&basepath, "%s/%s/receiving-from-uid-%d.pack",
 	    got_repo_get_path(gotd_session.repo), GOT_OBJECTS_PACK_DIR,
@@ -1344,13 +1380,8 @@ recv_packfile(struct gotd_session_client *client)
 		goto done;
 	}
 	packfd = -1;
-
 done:
 	free(basepath);
-	if (pipe[0] != -1 && close(pipe[0]) == -1 && err == NULL)
-		err = got_error_from_errno("close");
-	if (pipe[1] != -1 && close(pipe[1]) == -1 && err == NULL)
-		err = got_error_from_errno("close");
 	if (packfd != -1 && close(packfd) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	if (idxfd != -1 && close(idxfd) == -1 && err == NULL)
@@ -1510,6 +1541,14 @@ session_dispatch_client(int fd, short events, void *arg)
 				break;
 			}
 			break;
+		case GOTD_IMSG_PACKFILE_READY:
+			if (gotd_session.state != GOTD_STATE_EXPECT_PACKFILE ||
+			    gotd_session.repo_child_packfd == -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = send_packfds_to_repo_child();
+			break;
 		default:
 			log_debug("unexpected imsg %d", imsg.hdr.type);
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -1641,6 +1680,14 @@ session_dispatch_notifier(int fd, short event, void *arg)
 				log_warn("unexpected imsg %d", imsg.hdr.type);
 				break;
 			}
+
+			/* First notification on our list has now been sent. */
+			notif = STAILQ_FIRST(&notifications);
+			STAILQ_REMOVE_HEAD(&notifications, entry);
+			if (notif->fd != -1)
+				close(notif->fd);
+			free(notif);
+
 			notif = STAILQ_FIRST(&notifications);
 			if (notif == NULL) {
 				disconnect(client);
@@ -1986,6 +2033,7 @@ session_write_main(const char *title, const char *repo_path,
 	RB_INIT(&gotd_session.notification_refs);
 	RB_INIT(&gotd_session.notification_ref_namespaces);
 	STAILQ_INIT(&gotd_session.notification_targets);
+	gotd_session.repo_child_packfd = -1;
 
 	if (imsgbuf_init(&gotd_session.notifier_iev.ibuf, -1) == -1) {
 		err = got_error_from_errno("imsgbuf_init");
