@@ -1,7 +1,6 @@
 /*
  * Copyright (c) 2016, 2019, 2020-2022 Tracey Emery <tracey@traceyemery.net>
  * Copyright (c) 2015 Mike Larkin <mlarkin@openbsd.org>
- * Copyright (c) 2014 Reyk Floeter <reyk@openbsd.org>
  * Copyright (c) 2013 David Gwynne <dlg@openbsd.org>
  * Copyright (c) 2013 Florian Obser <florian@openbsd.org>
  *
@@ -25,7 +24,6 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
-#include <ctype.h>
 #include <dirent.h>
 #include <errno.h>
 #include <event.h>
@@ -54,38 +52,6 @@
 #include "log.h"
 #include "tmpl.h"
 
-static const struct querystring_keys querystring_keys[] = {
-	{ "action",		ACTION },
-	{ "commit",		COMMIT },
-	{ "file",		RFILE },
-	{ "folder",		FOLDER },
-	{ "headref",		HEADREF },
-	{ "index_page",		INDEX_PAGE },
-	{ "path",		PATH },
-};
-
-static const struct action_keys action_keys[] = {
-	{ "blame",	BLAME },
-	{ "blob",	BLOB },
-	{ "blobraw",	BLOBRAW },
-	{ "briefs",	BRIEFS },
-	{ "commits",	COMMITS },
-	{ "diff",	DIFF },
-	{ "error",	ERR },
-	{ "index",	INDEX },
-	{ "patch",	PATCH },
-	{ "summary",	SUMMARY },
-	{ "tag",	TAG },
-	{ "tags",	TAGS },
-	{ "tree",	TREE },
-	{ "rss",	RSS },
-};
-
-static const struct got_error *gotweb_init_querystring(struct querystring **);
-static const struct got_error *gotweb_parse_querystring(struct querystring *,
-    char *);
-static const struct got_error *gotweb_assign_querystring(struct querystring *,
-    char *, char *);
 static int gotweb_render_index(struct template *);
 static const struct got_error *gotweb_load_got_path(struct repo_dir **,
     const char *, struct request *);
@@ -96,12 +62,9 @@ static const struct got_error *gotweb_get_repo_description(char **,
 static const struct got_error *gotweb_get_clone_url(char **, struct server *,
     const char *, int);
 
-static void gotweb_free_querystring(struct querystring *);
 static void gotweb_free_repo_dir(struct repo_dir *);
 
-struct server *gotweb_get_server(const char *);
-
-static int
+int
 gotweb_reply(struct request *c, int status, const char *ctype,
     struct gotweb_url *location)
 {
@@ -112,7 +75,7 @@ gotweb_reply(struct request *c, int status, const char *ctype,
 
 	if (location) {
 		if (tp_writes(c->tp, "Location: ") == -1 ||
-		    gotweb_render_url(c, location) == -1 ||
+		    gotweb_render_absolute_url(c, location) == -1 ||
 		    tp_writes(c->tp, "\r\n") == -1)
 			return -1;
 	}
@@ -141,20 +104,6 @@ gotweb_reply_file(struct request *c, const char *ctype, const char *file,
 	return gotweb_reply(c, 200, ctype, NULL);
 }
 
-static void
-free_request(struct request *c)
-{
-	if (c->fd != -1)
-		close(c->fd);
-	if (c->tp != NULL)
-		template_free(c->tp);
-	if (c->t != NULL)
-		gotweb_free_transport(c->t);
-	free(c->buf);
-	free(c->outbuf);
-	free(c);
-}
-
 static struct socket *
 gotweb_get_socket(int sock_id)
 {
@@ -168,6 +117,18 @@ gotweb_get_socket(int sock_id)
 	return NULL;
 }
 
+static void
+cleanup_request(struct request *c)
+{
+	uint32_t request_id = c->request_id;
+
+	fcgi_cleanup_request(c);
+
+	if (imsg_compose_event(gotwebd_env->iev_auth, GOTWEBD_IMSG_REQ_ABORT,
+	    GOTWEBD_PROC_GOTWEB, -1, -1, &request_id, sizeof(request_id)) == -1)
+		log_warn("imsg_compose_event");
+}
+
 static struct request *
 recv_request(struct imsg *imsg)
 {
@@ -223,14 +184,14 @@ recv_request(struct imsg *imsg)
 	c->tp = template(c, fcgi_write, c->outbuf, GOTWEBD_CACHESIZE);
 	if (c->tp == NULL) {
 		log_warn("gotweb init template");
-		free_request(c);
+		cleanup_request(c);
 		return NULL;
 	}
 
 	c->sock = gotweb_get_socket(c->sock_id);
 	if (c->sock == NULL) {
 		log_warn("socket id '%d' not found", c->sock_id);
-		free_request(c);
+		cleanup_request(c);
 		return NULL;
 	}
 
@@ -238,15 +199,15 @@ recv_request(struct imsg *imsg)
 	error = gotweb_init_transport(&c->t);
 	if (error) {
 		log_warnx("gotweb init transport: %s", error->msg);
-		free_request(c);
+		cleanup_request(c);
 		return NULL;
 	}
 
 	/* get the gotwebd server */
-	srv = gotweb_get_server(c->server_name);
+	srv = gotweb_get_server(c->fcgi_params.server_name);
 	if (srv == NULL) {
-		log_warnx("server '%s' not found", c->server_name);
-		free_request(c);
+		log_warnx("server '%s' not found", c->fcgi_params.server_name);
+		cleanup_request(c);
 		return NULL;
 	}
 	c->srv = srv;
@@ -267,48 +228,43 @@ gotweb_process_request(struct request *c)
 	size_t len;
 	int r, binary = 0;
 
-	/* parse our querystring */
-	error = gotweb_init_querystring(&qs);
-	if (error) {
-		log_warnx("%s: %s", __func__, error->msg);
-		goto err;
-	}
+	/* querystring */
+	qs = &c->fcgi_params.qs;
 	c->t->qs = qs;
-	error = gotweb_parse_querystring(qs, c->querystring);
-	if (error) {
-		log_warnx("%s: %s", __func__, error->msg);
-		goto err;
-	}
 
 	/* Log the request. */
 	if (gotwebd_env->gotwebd_verbose > 0) {
+		struct gotwebd_fcgi_params *p = &c->fcgi_params;
 		char *server_name = NULL;
-		char *querystring = NULL;
 		char *document_uri = NULL;
+		const char *action_name = NULL;
 
-		if (c->server_name[0] &&
-		    stravis(&server_name, c->server_name, VIS_SAFE) == -1) {
+		if (p->server_name[0] &&
+		    stravis(&server_name, p->server_name, VIS_SAFE) == -1) {
 			log_warn("stravis");
 			server_name = NULL;
 		}
-		if (c->querystring[0] &&
-		    stravis(&querystring, c->querystring, VIS_SAFE) == -1) {
-			log_warn("stravis");
-			querystring = NULL;
-		}
-		if (c->document_uri[0] &&
-		    stravis(&document_uri, c->document_uri, VIS_SAFE) == -1) {
+
+		if (p->document_uri[0] &&
+		    stravis(&document_uri, p->document_uri, VIS_SAFE) == -1) {
 			log_warn("stravis");
 			document_uri = NULL;
 		}
 
-		log_info("processing request: server='%s' query='%s' "
-		    "document_uri='%s'",
+		action_name = gotweb_action_name(qs->action);
+		log_info("processing request: server='%s' action='%s' "
+		    "commit='%s', file='%s', folder='%s', headref='%s' "
+		    "index_page=%d path='%s' document_uri='%s'",
 		    server_name ? server_name : "",
-		    querystring ? querystring : "",
+		    action_name ? action_name : "",
+		    qs->commit,
+		    qs->file,
+		    qs->folder,
+		    qs->headref,
+		    qs->index_page,
+		    qs->path,
 		    document_uri ? document_uri : "");
 		free(server_name);
-		free(querystring);
 		free(document_uri);
 	}
 
@@ -321,14 +277,14 @@ gotweb_process_request(struct request *c)
 	if (qs->action == BLAME || qs->action == BLOB ||
 	    qs->action == BLOBRAW || qs->action == DIFF ||
 	    qs->action == PATCH) {
-		if (qs->commit == NULL) {
+		if (qs->commit[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
 	}
 
 	if (qs->action != INDEX) {
-		if (qs->path == NULL) {
+		if (qs->path[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
@@ -340,7 +296,7 @@ gotweb_process_request(struct request *c)
 	}
 
 	if (qs->action == BLOBRAW || qs->action == BLOB) {
-		if (qs->folder == NULL || qs->file == NULL) {
+		if (qs->folder[0] == '\0' || qs->file[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
@@ -357,7 +313,7 @@ gotweb_process_request(struct request *c)
 
 	switch (qs->action) {
 	case BLAME:
-		if (qs->folder == NULL || qs->file == NULL) {
+		if (qs->folder[0] == '\0' || qs->file[0] == '\0') {
 			error = got_error(GOT_ERR_BAD_QUERYSTRING);
 			goto err;
 		}
@@ -374,10 +330,10 @@ gotweb_process_request(struct request *c)
 			struct gotweb_url url = {
 				.index_page = -1,
 				.action = BLOBRAW,
-				.path = qs->path,
-				.commit = qs->commit,
-				.folder = qs->folder,
-				.file = qs->file,
+				.path = qs->path[0] ? qs->path : NULL,
+				.commit = qs->commit[0] ? qs->commit : NULL,
+				.folder = qs->folder[0] ? qs->folder : NULL,
+				.file = qs->file[0] ? qs->file : NULL,
 			};
 
 			return gotweb_reply(c, 302, NULL, &url);
@@ -492,11 +448,12 @@ gotweb_process_request(struct request *c)
 		}
 		qs->action = SUMMARY;
 		commit = TAILQ_FIRST(&c->t->repo_commits);
-		if (commit && qs->commit == NULL) {
-			qs->commit = strdup(commit->commit_id);
-			if (qs->commit == NULL) {
-				error = got_error_from_errno("strdup");
-				log_warn("%s: strdup", __func__);
+		if (commit && qs->commit[0] == '\0') {
+			if (strlcpy(qs->commit, commit->commit_id,
+			    sizeof(qs->commit)) >= sizeof(qs->commit)) {
+				error = got_error_msg(GOT_ERR_NO_SPACE,
+				    "commit ID too long");
+				log_warn("%s: %s", __func__, error->msg);
 				goto err;
 			}
 		}
@@ -580,209 +537,26 @@ gotweb_init_transport(struct transport **t)
 	return error;
 }
 
-static const struct got_error *
-gotweb_init_querystring(struct querystring **qs)
+struct gotwebd_repo *
+gotweb_get_repository(struct server *server, const char *name)
 {
-	const struct got_error *error = NULL;
+	struct gotwebd_repo *repo;
 
-	*qs = calloc(1, sizeof(**qs));
-	if (*qs == NULL)
-		return got_error_from_errno2(__func__, "calloc");
+	TAILQ_FOREACH(repo, &server->repos, entry) {
+		if (strncmp(repo->name, name, strlen(repo->name)) != 0)
+			continue;
+	
+		if (strlen(name) == strlen(repo->name))
+			return repo;
 
-	(*qs)->headref = strdup("HEAD");
-	if ((*qs)->headref == NULL) {
-		free(*qs);
-		*qs = NULL;
-		return got_error_from_errno2(__func__, "strdup");
-	}
-
-	(*qs)->action = INDEX;
-
-	return error;
-}
-
-static const struct got_error *
-gotweb_parse_querystring(struct querystring *qs, char *qst)
-{
-	const struct got_error *error = NULL;
-	char *tok1 = NULL, *tok1_pair = NULL, *tok1_end = NULL;
-	char *tok2 = NULL, *tok2_pair = NULL, *tok2_end = NULL;
-
-	if (qst == NULL)
-		return error;
-
-	tok1 = strdup(qst);
-	if (tok1 == NULL)
-		return got_error_from_errno2(__func__, "strdup");
-
-	tok1_pair = tok1;
-	tok1_end = tok1;
-
-	while (tok1_pair != NULL) {
-		strsep(&tok1_end, "&");
-
-		tok2 = strdup(tok1_pair);
-		if (tok2 == NULL) {
-			free(tok1);
-			return got_error_from_errno2(__func__, "strdup");
-		}
-
-		tok2_pair = tok2;
-		tok2_end = tok2;
-
-		while (tok2_pair != NULL) {
-			strsep(&tok2_end, "=");
-			if (tok2_end) {
-				error = gotweb_assign_querystring(qs, tok2_pair,
-				    tok2_end);
-				if (error)
-					goto err;
-			}
-			tok2_pair = tok2_end;
-		}
-		free(tok2);
-		tok1_pair = tok1_end;
-	}
-	free(tok1);
-	return error;
-err:
-	free(tok2);
-	free(tok1);
-	return error;
-}
-
-/*
- * Adapted from usr.sbin/httpd/httpd.c url_decode.
- */
-static const struct got_error *
-gotweb_urldecode(char *url)
-{
-	char		*p, *q;
-	char		 hex[3];
-	unsigned long	 x;
-
-	hex[2] = '\0';
-	p = q = url;
-
-	while (*p != '\0') {
-		switch (*p) {
-		case '%':
-			/* Encoding character is followed by two hex chars */
-			if (!isxdigit((unsigned char)p[1]) ||
-			    !isxdigit((unsigned char)p[2]) ||
-			    (p[1] == '0' && p[2] == '0'))
-				return got_error(GOT_ERR_BAD_QUERYSTRING);
-
-			hex[0] = p[1];
-			hex[1] = p[2];
-
-			/*
-			 * We don't have to validate "hex" because it is
-			 * guaranteed to include two hex chars followed by nul.
-			 */
-			x = strtoul(hex, NULL, 16);
-			*q = (char)x;
-			p += 2;
-			break;
-		default:
-			*q = *p;
-			break;
-		}
-		p++;
-		q++;
-	}
-	*q = '\0';
-
-	return NULL;
-}
-
-static const struct got_error *
-gotweb_assign_querystring(struct querystring *qs, char *key, char *value)
-{
-	const struct got_error *error = NULL;
-	const char *errstr;
-	int a_cnt, el_cnt;
-
-	error = gotweb_urldecode(value);
-	if (error)
-		return error;
-
-	for (el_cnt = 0; el_cnt < nitems(querystring_keys); el_cnt++) {
-		if (strcmp(key, querystring_keys[el_cnt].name) != 0)
+		if (strlen(name) != strlen(repo->name) + 4)
 			continue;
 
-		switch (querystring_keys[el_cnt].element) {
-		case ACTION:
-			for (a_cnt = 0; a_cnt < nitems(action_keys); a_cnt++) {
-				if (strcmp(value, action_keys[a_cnt].name) != 0)
-					continue;
-				qs->action = action_keys[a_cnt].action;
-				goto qa_found;
-			}
-			qs->action = ERR;
-qa_found:
-			break;
-		case COMMIT:
-			qs->commit = strdup(value);
-			if (qs->commit == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case RFILE:
-			qs->file = strdup(value);
-			if (qs->file == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case FOLDER:
-			qs->folder = strdup(value);
-			if (qs->folder == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case HEADREF:
-			free(qs->headref);
-			qs->headref = strdup(value);
-			if (qs->headref == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		case INDEX_PAGE:
-			if (*value == '\0')
-				break;
-			qs->index_page = strtonum(value, INT64_MIN,
-			    INT64_MAX, &errstr);
-			if (errstr) {
-				error = got_error_from_errno3(__func__,
-				    "strtonum", errstr);
-				goto done;
-			}
-			if (qs->index_page < 0)
-				qs->index_page = 0;
-			break;
-		case PATH:
-			qs->path = strdup(value);
-			if (qs->path == NULL) {
-				error = got_error_from_errno2(__func__,
-				    "strdup");
-				goto done;
-			}
-			break;
-		}
-
-		/* entry found */
-		break;
+		if (strcmp(name + strlen(repo->name), ".git") == 0)
+			return repo;
 	}
-done:
-	return error;
+
+	return NULL;
 }
 
 void
@@ -814,19 +588,6 @@ gotweb_free_repo_commit(struct repo_commit *rc)
 	free(rc);
 }
 
-static void
-gotweb_free_querystring(struct querystring *qs)
-{
-	if (qs != NULL) {
-		free(qs->commit);
-		free(qs->file);
-		free(qs->folder);
-		free(qs->headref);
-		free(qs->path);
-	}
-	free(qs);
-}
-
 static void
 gotweb_free_repo_dir(struct repo_dir *repo_dir)
 {
@@ -858,7 +619,7 @@ gotweb_free_transport(struct transport *t)
 		gotweb_free_repo_tag(rt);
 	}
 	gotweb_free_repo_dir(t->repo_dir);
-	gotweb_free_querystring(t->qs);
+	t->qs = NULL;
 	free(t->more_id);
 	free(t->tags_more_id);
 	if (t->blob)
@@ -886,7 +647,7 @@ gotweb_index_navs(struct request *c, struct gotweb_url *prev, int *have_prev,
     struct gotweb_url *next, int *have_next)
 {
 	struct transport *t = c->t;
-	struct querystring *qs = t->qs;
+	const struct querystring *qs = t->qs;
 	struct server *srv = c->srv;
 
 	*have_prev = *have_next = 0;
@@ -916,7 +677,7 @@ gotweb_render_index(struct template *tp)
 	struct request *c = tp->tp_arg;
 	struct server *srv = c->srv;
 	struct transport *t = c->t;
-	struct querystring *qs = t->qs;
+	const struct querystring *qs = t->qs;
 	struct repo_dir *repo_dir = NULL;
 	struct dirent **sd_dent = t->repos;
 	unsigned int d_i, d_disp = 0;
@@ -1061,6 +822,8 @@ const char *
 gotweb_action_name(int action)
 {
 	switch (action) {
+	case NO_ACTION:
+		return "no action";
 	case BLAME:
 		return "blame";
 	case BLOB:
@@ -1126,10 +889,14 @@ gotweb_render_url(struct request *c, struct gotweb_url *url)
 	}
 
 	if (url->folder) {
-		tmp = gotweb_urlencode(url->folder);
-		if (tmp == NULL)
-			return -1;
-		r = tp_writef(c->tp, "%sfolder=%s", sep, tmp);
+		if (got_path_is_root_dir(url->folder))
+			tmp = NULL;
+		else {
+			tmp = gotweb_urlencode(url->folder);
+			if (tmp == NULL)
+				return -1;
+		}
+		r = tp_writef(c->tp, "%sfolder=%s", sep, tmp ? tmp : "");
 		free(tmp);
 		if (r == -1)
 			return -1;
@@ -1172,27 +939,87 @@ int
 gotweb_render_absolute_url(struct request *c, struct gotweb_url *url)
 {
 	struct template	*tp = c->tp;
-	const char	*proto = c->https ? "https" : "http";
+	struct gotwebd_fcgi_params *p = &c->fcgi_params;
+	const char	*proto = p->https ? "https" : "http";
 
 	if (tp_writes(tp, proto) == -1 ||
 	    tp_writes(tp, "://") == -1 ||
-	    tp_htmlescape(tp, c->server_name) == -1 ||
-	    tp_htmlescape(tp, c->document_uri) == -1)
+	    tp_htmlescape(tp, p->server_name) == -1 ||
+	    tp_htmlescape(tp, p->document_uri) == -1)
 		return -1;
 
 	return gotweb_render_url(c, url);
 }
 
+/* 
+ * Ensure that a path sent in the request will not escape from the given
+ * parent directory. This matters for got-portable where we are not
+ * necessarily running in chroot and cannot be protected by unveil(2).
+ */
+static const struct got_error *
+validate_path(const char *path, const char *parent_path,
+    const char *orig_path)
+{
+	const struct got_error *error = NULL;
+	char *abspath;
+
+	abspath = realpath(path, NULL);
+	if (abspath == NULL) {
+		/* Paths which cannot be resolved are safe. */
+		if (errno == ENOENT || errno == EACCES || errno == ENOTDIR)
+			return NULL;
+		return got_error_from_errno("realpath");
+	}
+
+	if (!got_path_is_child(abspath, parent_path, strlen(parent_path)))
+		error = got_error_path(orig_path, GOT_ERR_NOT_GIT_REPO);
+
+	free(abspath);
+	return error;
+}
+
+static enum gotwebd_access
+auth_check(struct request *c, struct gotwebd_access_rule_list *rules)
+{
+	struct gotwebd *env = gotwebd_env;
+	enum gotwebd_access access = GOTWEBD_ACCESS_NO_MATCH;
+	struct gotwebd_access_rule *rule;
+
+	/*
+	 * The www user ID implies that no user authentication occurred.
+	 * But authentication is enabled so we must deny this request.
+	 */
+	if (c->client_uid == env->www_uid)
+		return GOTWEBD_ACCESS_DENIED;
+
+	/*
+	 * We cannot access /etc/passwd in this process so we cannot
+	 * verify the client's user ID outselves here.
+	 * Match rules against the access identifier which has already
+	 * passed authentication in the auth process.
+	 */
+	STAILQ_FOREACH(rule, rules, entry) {
+		if (strcmp(rule->identifier, c->access_identifier) == 0)
+			access = rule->access;
+	}
+
+	return access;
+}
+
 static const struct got_error *
 gotweb_load_got_path(struct repo_dir **rp, const char *dir,
     struct request *c)
 {
 	const struct got_error *error = NULL;
+	struct gotwebd *env = gotwebd_env;
 	struct server *srv = c->srv;
 	struct transport *t = c->t;
 	struct repo_dir *repo_dir;
 	DIR *dt;
 	char *dir_test;
+	struct gotwebd_repo *repo;
+	enum gotwebd_auth_config auth_config = 0;
+	enum gotwebd_access access = GOTWEBD_ACCESS_DENIED;
 
 	*rp = calloc(1, sizeof(**rp));
 	if (*rp == NULL)
@@ -1203,6 +1030,12 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 	    GOTWEB_GIT_DIR) == -1)
 		return got_error_from_errno("asprintf");
 
+	error = validate_path(dir_test, srv->repos_path, dir);
+	if (error) {
+		free(dir_test);
+		return error;
+	}
+
 	dt = opendir(dir_test);
 	if (dt == NULL) {
 		free(dir_test);
@@ -1226,9 +1059,51 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 	repo_dir->path = dir_test;
 	dir_test = NULL;
 
-	repo_dir->name = strdup(repo_dir->path + strlen(srv->repos_path) + 1);
-	if (repo_dir->name == NULL) {
-		error = got_error_from_errno("strdup");
+	error = got_path_basename(&repo_dir->name, repo_dir->path);
+	if (error)
+		goto err;
+
+	repo = gotweb_get_repository(srv, repo_dir->name);
+	if (repo) {
+		auth_config = repo->auth_config;
+		switch (auth_config) {
+		case GOTWEBD_AUTH_DISABLED:
+			access = GOTWEBD_ACCESS_PERMITTED;
+			break;
+		case GOTWEBD_AUTH_SECURE:
+		case GOTWEBD_AUTH_INSECURE:
+			access = auth_check(c, &repo->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = auth_check(c, &srv->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = auth_check(c, &env->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = GOTWEBD_ACCESS_DENIED;
+			break;
+		}
+	} else {
+		auth_config = srv->auth_config;
+		switch (auth_config) {
+		case GOTWEBD_AUTH_DISABLED:
+			access = GOTWEBD_ACCESS_PERMITTED;
+			break;
+		case GOTWEBD_AUTH_SECURE:
+		case GOTWEBD_AUTH_INSECURE:
+			access = auth_check(c, &srv->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = auth_check(c, &env->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = GOTWEBD_ACCESS_DENIED;
+			break;
+		}
+	}
+
+	if (access != GOTWEBD_ACCESS_PERMITTED &&
+	    access != GOTWEBD_ACCESS_DENIED)
+		fatalx("invalid access check result %d", access);
+
+	if (access != GOTWEBD_ACCESS_PERMITTED) {
+		error = got_error_path(repo_dir->name, GOT_ERR_NOT_GIT_REPO);
 		goto err;
 	}
 
@@ -1389,21 +1264,39 @@ gotweb_shutdown(void)
 {
 	imsgbuf_clear(&gotwebd_env->iev_parent->ibuf);
 	free(gotwebd_env->iev_parent);
-	if (gotwebd_env->iev_server) {
-		imsgbuf_clear(&gotwebd_env->iev_server->ibuf);
-		free(gotwebd_env->iev_server);
-	}
 
-	sockets_purge(gotwebd_env);
+	imsgbuf_clear(&gotwebd_env->iev_auth->ibuf);
+	free(gotwebd_env->iev_auth);
+
+	config_free_access_rules(&gotwebd_env->access_rules);
 
 	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
 		struct server *srv;
 
 		srv = TAILQ_FIRST(&gotwebd_env->servers);
 		TAILQ_REMOVE(&gotwebd_env->servers, srv, entry);
+
+		config_free_access_rules(&srv->access_rules);
+		config_free_repos(&srv->repos);
 		free(srv);
 	}
 
+	while (!TAILQ_EMPTY(&gotwebd_env->sockets)) {
+		struct socket *sock;
+
+		sock = TAILQ_FIRST(&gotwebd_env->sockets);
+		TAILQ_REMOVE(&gotwebd_env->sockets, sock, entry);
+		free(sock);
+	}
+
+	while (!TAILQ_EMPTY(&gotwebd_env->addresses)) {
+		struct address *h;
+
+		h = TAILQ_FIRST(&gotwebd_env->addresses);
+		TAILQ_REMOVE(&gotwebd_env->addresses, h, entry);
+		free(h);
+	}
+
 	free(gotwebd_env);
 
 	exit(0);
@@ -1439,8 +1332,8 @@ gotweb_launch(struct gotwebd *env)
 	struct server *srv;
 	const struct got_error *error;
 
-	if (env->iev_server == NULL)
-		fatal("server process not connected");
+	if (env->iev_auth == NULL)
+		fatal("auth process not connected");
 
 #ifndef PROFILE
 	if (pledge("stdio rpath recvfd sendfd proc exec unveil", NULL) == -1)
@@ -1459,18 +1352,7 @@ gotweb_launch(struct gotwebd *env)
 	if (unveil(NULL, NULL) == -1)
 		fatal("unveil");
 
-	event_add(&env->iev_server->ev, NULL);
-}
-
-static void
-send_request_done(struct imsgev *iev, int request_id)
-{
-	struct gotwebd		*env = gotwebd_env;
-
-	if (imsg_compose_event(env->iev_server, GOTWEBD_IMSG_REQ_DONE,
-	    GOTWEBD_PROC_GOTWEB, getpid(), -1,
-	    &request_id, sizeof(request_id)) == -1)
-		log_warn("imsg_compose_event");
+	event_add(&env->iev_auth->ev, NULL);
 }
 
 static void
@@ -1516,8 +1398,9 @@ gotweb_dispatch_server(int fd, short event, void *arg)
 						    request_id);
 					}
 				}
-				free_request(c);
-				send_request_done(iev, request_id);
+
+				fcgi_create_end_record(c);
+				cleanup_request(c);
 			}
 			break;
 		default:
@@ -1538,24 +1421,21 @@ gotweb_dispatch_server(int fd, short event, void *arg)
 }
 
 static void
-recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
+recv_auth_pipe(struct gotwebd *env, struct imsg *imsg)
 {
 	struct imsgev *iev;
 	int fd;
 
-	if (env->iev_server != NULL) {
-		log_warn("server pipe already received"); 
-		return;
-	}
+	if (env->iev_auth != NULL)
+		fatalx("auth process already connected");
 
 	fd = imsg_get_fd(imsg);
 	if (fd == -1)
-		fatalx("invalid server pipe fd");
+		fatalx("invalid auth pipe fd");
 
 	iev = calloc(1, sizeof(*iev));
 	if (iev == NULL)
 		fatal("calloc");
-
 	if (imsgbuf_init(&iev->ibuf, fd) == -1)
 		fatal("imsgbuf_init");
 	imsgbuf_allow_fdpass(&iev->ibuf);
@@ -1565,7 +1445,7 @@ recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
 	event_set(&iev->ev, fd, EV_READ, gotweb_dispatch_server, iev);
 	imsg_event_add(iev);
 
-	env->iev_server = iev;
+	env->iev_auth = iev;
 }
 
 static void
@@ -1575,6 +1455,8 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 	struct imsgbuf		*ibuf;
 	struct imsg		 imsg;
 	struct gotwebd		*env = gotwebd_env;
+	struct server		*srv;
+	struct gotwebd_repo	*repo;
 	ssize_t			 n;
 	int			 shut = 0;
 
@@ -1598,9 +1480,35 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 			break;
 
 		switch (imsg.hdr.type) {
+		case GOTWEBD_IMSG_CFG_ACCESS_RULE:
+			if (TAILQ_EMPTY(&env->servers)) {
+				/* global access rule */
+				config_get_access_rule(&env->access_rules,
+				    &imsg);
+			} else {
+				srv = TAILQ_LAST(&env->servers, serverlist);
+				if (TAILQ_EMPTY(&srv->repos)) {
+					/* per-server access rule */
+					config_get_access_rule(
+					    &srv->access_rules, &imsg);
+				} else {
+					/* per-repository access rule */
+					repo = TAILQ_LAST(&srv->repos,
+					    gotwebd_repolist);
+					config_get_access_rule(
+					    &repo->access_rules, &imsg);
+				}
+			}
+			break;
 		case GOTWEBD_IMSG_CFG_SRV:
 			config_getserver(env, &imsg);
 			break;
+		case GOTWEBD_IMSG_CFG_REPO:
+			if (TAILQ_EMPTY(&env->servers))
+				fatalx("%s: unexpected CFG_REPO msg", __func__);
+			srv = TAILQ_LAST(&env->servers, serverlist);
+			config_get_repository(&srv->repos, &imsg);
+			break;
 		case GOTWEBD_IMSG_CFG_FD:
 			config_getfd(env, &imsg);
 			break;
@@ -1611,7 +1519,17 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 			config_getcfg(env, &imsg);
 			break;
 		case GOTWEBD_IMSG_CTL_PIPE:
-			recv_server_pipe(env, &imsg);
+			recv_auth_pipe(env, &imsg);
+			break;
+		case GOTWEBD_IMSG_AUTH_CONF:
+			if (imsg_get_data(&imsg, &env->auth_config,
+			    sizeof(env->auth_config)) == -1)
+				fatalx("%s: invalid AUTH_CONF msg", __func__);
+			break;
+		case GOTWEBD_IMSG_WWW_UID:
+			if (imsg_get_data(&imsg, &env->www_uid,
+			    sizeof(env->www_uid)) == -1)
+				fatalx("%s: invalid WWW_UID msg", __func__);
 			break;
 		case GOTWEBD_IMSG_CTL_START:
 			gotweb_launch(env);
@@ -1643,8 +1561,9 @@ gotweb(struct gotwebd *env, int fd)
 
 	sockets_rlimit(-1);
 
-	if ((env->iev_parent = malloc(sizeof(*env->iev_parent))) == NULL)
-		fatal("malloc");
+	env->iev_parent = calloc(1, sizeof(*env->iev_parent));
+	if (env->iev_parent == NULL)
+		fatal("calloc");
 	if (imsgbuf_init(&env->iev_parent->ibuf, fd) == -1)
 		fatal("imsgbuf_init");
 	imsgbuf_allow_fdpass(&env->iev_parent->ibuf);
