@@ -39,6 +39,10 @@
 #define GOTWEBD_WWW_USER	 "www"
 #endif
 
+#define GOTWEBD_LOGIN_CMD	"weblogin"
+#define GOTWEBD_LOGIN_SOCKET	 "/var/run/gotweb-login.sock"
+#define GOTWEBD_LOGIN_TIMEOUT	 300 /* in seconds */
+
 #define GOTWEBD_MAXDESCRSZ	 1024
 #define GOTWEBD_MAXCLONEURLSZ	 1024
 #define GOTWEBD_CACHESIZE	 1024
@@ -55,6 +59,8 @@
 #define MAX_QUERYSTRING		 2048
 #define MAX_DOCUMENT_URI	 255
 #define MAX_SERVER_NAME		 255
+#define MAX_AUTH_COOKIE		 255
+#define MAX_IDENTIFIER_SIZE	 32
 
 #define GOTWEB_GIT_DIR		 ".git"
 
@@ -94,7 +100,6 @@
     (((n) + (FCGI_ALIGNMENT - 1)) & ~(FCGI_ALIGNMENT - 1))
 
 #define FD_RESERVE		 5
-#define FD_NEEDED		 6
 
 #define FCGI_BEGIN_REQUEST	 1
 #define FCGI_ABORT_REQUEST	 2
@@ -123,7 +128,10 @@ struct got_reflist_head;
 
 enum gotwebd_proc_type {
 	GOTWEBD_PROC_PARENT,
-	GOTWEBD_PROC_SERVER,
+	GOTWEBD_PROC_SOCKETS,
+	GOTWEBD_PROC_FCGI,
+	GOTWEBD_PROC_LOGIN,
+	GOTWEBD_PROC_AUTH,
 	GOTWEBD_PROC_GOTWEB,
 };
 
@@ -131,11 +139,19 @@ enum imsg_type {
 	GOTWEBD_IMSG_CFG_SRV,
 	GOTWEBD_IMSG_CFG_SOCK,
 	GOTWEBD_IMSG_CFG_FD,
+	GOTWEBD_IMSG_CFG_ACCESS_RULE,
+	GOTWEBD_IMSG_CFG_REPO,
 	GOTWEBD_IMSG_CFG_DONE,
 	GOTWEBD_IMSG_CTL_PIPE,
 	GOTWEBD_IMSG_CTL_START,
+	GOTWEBD_IMSG_LOGIN_SECRET,
+	GOTWEBD_IMSG_AUTH_SECRET,
+	GOTWEBD_IMSG_AUTH_CONF,
+	GOTWEBD_IMSG_FCGI_PARSE_PARAMS,
+	GOTWEBD_IMSG_FCGI_PARAMS,
+	GOTWEBD_IMSG_WWW_UID,
+	GOTWEBD_IMSG_REQ_ABORT,
 	GOTWEBD_IMSG_REQ_PROCESS,
-	GOTWEBD_IMSG_REQ_DONE,
 };
 
 struct imsgev {
@@ -209,7 +225,7 @@ struct transport {
 	struct got_reflist_head	 refs;
 	struct got_repository	*repo;
 	struct repo_dir		*repo_dir;
-	struct querystring	*qs;
+	const struct querystring *qs;
 	char			*more_id;
 	char			*tags_more_id;
 	unsigned int		 repos_total;
@@ -240,6 +256,50 @@ enum socket_priv_fds {
 	PRIV_FDS__MAX,
 };
 
+struct gotwebd_fcgi_record {
+	uint32_t			 request_id;
+	uint8_t				 record[FCGI_RECORD_SIZE];
+	size_t				 record_len;
+};
+
+enum query_actions {
+	NO_ACTION = 0,
+	BLAME,
+	BLOB,
+	BLOBRAW,
+	BRIEFS,
+	COMMITS,
+	DIFF,
+	ERR,
+	INDEX,
+	PATCH,
+	SUMMARY,
+	TAG,
+	TAGS,
+	TREE,
+	RSS,
+};
+
+struct querystring {
+	enum query_actions action;
+	char		 commit[GOT_OBJECT_ID_HEX_MAXLEN];
+	char		 file[NAME_MAX];
+	char		 folder[PATH_MAX];
+	char		 headref[MAX_DOCUMENT_URI];
+	int		 index_page;
+	char		 path[PATH_MAX];
+	char		 login[MAX_AUTH_COOKIE];
+};
+
+struct gotwebd_fcgi_params {
+	uint32_t			 request_id;
+	struct querystring		 qs;
+	char				 document_uri[MAX_DOCUMENT_URI];
+	char				 server_name[MAX_SERVER_NAME];
+	char				 auth_cookie[MAX_AUTH_COOKIE];
+	int				 https;
+};
+
 struct template;
 struct request {
 	TAILQ_ENTRY(request)		entry;
@@ -253,24 +313,23 @@ struct request {
 	uint16_t			 id;
 	int				 fd;
 	int				 priv_fd[PRIV_FDS__MAX];
-	int				 resp_fd;
-	struct event			*resp_event;
 	int				 sock_id;
 	uint32_t			 request_id;
+	int				 worker_idx;
 
 	uint8_t				 *buf;
-	size_t				 buf_pos;
 	size_t				 buf_len;
 
 	uint8_t				 *outbuf;
 
-	char				 querystring[MAX_QUERYSTRING];
-	char				 document_uri[MAX_DOCUMENT_URI];
-	char				 server_name[MAX_SERVER_NAME];
-	int				 https;
+	struct gotwebd_fcgi_params	 fcgi_params;
+	int				 nparams;
+	int				 nparams_parsed;
 
-	uint8_t				 request_started;
 	int				 client_status;
+
+	uid_t				 client_uid;
+	char				 access_identifier[MAX_IDENTIFIER_SIZE];
 };
 TAILQ_HEAD(requestlist, request);
 
@@ -298,6 +357,36 @@ struct address {
 };
 TAILQ_HEAD(addresslist, address);
 
+enum gotwebd_auth_config {
+	GOTWEBD_AUTH_DISABLED	= 0xf00000ff,
+	GOTWEBD_AUTH_SECURE	= 0x00808000,
+	GOTWEBD_AUTH_INSECURE	= 0x0f7f7f00
+};
+
+enum gotwebd_access {
+	GOTWEBD_ACCESS_NO_MATCH = -2,
+	GOTWEBD_ACCESS_DENIED = -1,
+	GOTWEBD_ACCESS_PERMITTED = 1
+};
+
+struct gotwebd_access_rule {
+	STAILQ_ENTRY(gotwebd_access_rule) entry;
+
+	enum gotwebd_access access;
+	char identifier[MAX_IDENTIFIER_SIZE];
+};
+STAILQ_HEAD(gotwebd_access_rule_list, gotwebd_access_rule);
+
+struct gotwebd_repo {
+	TAILQ_ENTRY(gotwebd_repo)	 entry;
+
+	char name[NAME_MAX];
+
+	enum gotwebd_auth_config	auth_config;
+	struct gotwebd_access_rule_list access_rules;
+};
+TAILQ_HEAD(gotwebd_repolist, gotwebd_repo);
+
 struct server {
 	TAILQ_ENTRY(server)	 entry;
 
@@ -322,11 +411,19 @@ struct server {
 	int		 show_repo_description;
 	int		 show_repo_cloneurl;
 	int		 respect_exportok;
+
+	enum gotwebd_auth_config auth_config;
+	struct gotwebd_access_rule_list access_rules;
+
+	struct gotwebd_repolist	 repos;
 };
 TAILQ_HEAD(serverlist, server);
 
 enum client_action {
 	CLIENT_CONNECT,
+	CLIENT_FCGI_BEGIN,
+	CLIENT_FCGI_PARAMS,
+	CLIENT_FCGI_STDIN,
 	CLIENT_REQUEST,
 	CLIENT_DISCONNECT,
 };
@@ -357,6 +454,12 @@ struct gotwebd {
 	struct socketlist	sockets;
 	struct addresslist	addresses;
 
+	struct socket	*login_sock;
+	struct event	 login_pause_ev;
+
+	enum gotwebd_auth_config auth_config;
+	struct gotwebd_access_rule_list access_rules;
+
 	int		 pack_fds[GOTWEB_PACK_NUM_TEMPFILES];
 	int		 priv_fd[PRIV_FDS__MAX];
 
@@ -368,17 +471,20 @@ struct gotwebd {
 	int		 gotwebd_verbose;
 
 	struct imsgev	*iev_parent;
-	struct imsgev	*iev_server;
+	struct imsgev	*iev_sockets;
+	struct imsgev	*iev_fcgi;
+	struct imsgev	*iev_login;
+	struct imsgev	*iev_gotsh;
+	struct imsgev	*iev_auth;
 	struct imsgev	*iev_gotweb;
-	size_t		 nserver;
 
-	uint16_t	 prefork_gotwebd;
-	int		 servers_pending;
+	uint16_t	 prefork;
+	int		 auth_pending;
 	int		 gotweb_pending;
-
-	int		 server_cnt;
+	int		 *worker_load;
 
 	char		 httpd_chroot[PATH_MAX];
+	uid_t		 www_uid;
 };
 
 /*
@@ -395,16 +501,6 @@ struct gotweb_url {
 	const char	*path;
 };
 
-struct querystring {
-	uint8_t		 action;
-	char		*commit;
-	char		*file;
-	char		*folder;
-	char		*headref;
-	int		 index_page;
-	char		*path;
-};
-
 struct querystring_keys {
 	const char	*name;
 	int		 element;
@@ -423,23 +519,7 @@ enum querystring_elements {
 	HEADREF,
 	INDEX_PAGE,
 	PATH,
-};
-
-enum query_actions {
-	BLAME,
-	BLOB,
-	BLOBRAW,
-	BRIEFS,
-	COMMITS,
-	DIFF,
-	ERR,
-	INDEX,
-	PATCH,
-	SUMMARY,
-	TAG,
-	TAGS,
-	TREE,
-	RSS,
+	LOGIN,
 };
 
 extern struct gotwebd	*gotwebd_env;
@@ -453,8 +533,12 @@ int	 imsg_compose_event(struct imsgev *, uint16_t, uint32_t,
 	    pid_t, int, const void *, size_t);
 int	 main_compose_sockets(struct gotwebd *, uint32_t, int,
 	    const void *, uint16_t);
+int	 main_compose_login(struct gotwebd *, uint32_t, int,
+	    const void *, uint16_t);
 int	 sockets_compose_main(struct gotwebd *, uint32_t,
 	    const void *, uint16_t);
+int	 main_compose_auth(struct gotwebd *, uint32_t, int,
+	    const void *, uint16_t);
 int	 main_compose_gotweb(struct gotwebd *, uint32_t, int,
 	    const void *, uint16_t);
 
@@ -462,12 +546,26 @@ int	 main_compose_gotweb(struct gotwebd *, uint32_t, int,
 void sockets(struct gotwebd *, int);
 void sockets_parse_sockets(struct gotwebd *);
 void sockets_socket_accept(int, short, void *);
+struct socket *sockets_conf_new_socket(int, struct address *);
 int sockets_privinit(struct gotwebd *, struct socket *, uid_t, gid_t);
-void sockets_purge(struct gotwebd *);
 void sockets_rlimit(int);
-void sockets_del_request(struct request *);
+
+/* login.c */
+char *login_gen_token(uint64_t, const char *, time_t, const char *, size_t,
+    const char *);
+int login_check_token(uid_t *, char **, const char *, const char *, size_t,
+    const char *);
+int login_privinit(struct gotwebd *, uid_t, gid_t);
+void gotwebd_login(struct gotwebd *, int);
+
+/* auth.c */
+void gotwebd_auth(struct gotwebd *, int);
 
 /* gotweb.c */
+struct server *gotweb_get_server(const char *);
+struct gotwebd_repo * gotweb_get_repository(struct server *, const char *);
+int gotweb_reply(struct request *c, int status, const char *ctype,
+    struct gotweb_url *);
 void gotweb_index_navs(struct request *, struct gotweb_url *, int *,
     struct gotweb_url *, int *);
 int gotweb_render_age(struct template *, time_t);
@@ -499,17 +597,19 @@ int	gotweb_render_summary(struct template *);
 int	gotweb_render_blame(struct template *);
 int	gotweb_render_patch(struct template *);
 int	gotweb_render_rss(struct template *);
+int	gotweb_render_unauthorized(struct template *);
 
 /* parse.y */
+struct gotwebd_repo * gotwebd_new_repo(const char *);
 int parse_config(const char *, struct gotwebd *);
 int cmdline_symset(char *);
 
 /* fcgi.c */
-void fcgi_request(int, short, void *);
-void fcgi_timeout(int, short, void *);
+void fcgi_init_querystring(struct querystring *);
 void fcgi_cleanup_request(struct request *);
 void fcgi_create_end_record(struct request *);
 int fcgi_write(void *, const void *, size_t);
+void gotwebd_fcgi(struct gotwebd *, int);
 
 /* got_operations.c */
 const struct got_error *got_gotweb_closefile(FILE *);
@@ -537,4 +637,11 @@ int config_getsock(struct gotwebd *, struct imsg *);
 int config_setfd(struct gotwebd *);
 int config_getfd(struct gotwebd *, struct imsg *);
 int config_getcfg(struct gotwebd *, struct imsg *);
+void config_set_access_rules(struct imsgev *,
+    struct gotwebd_access_rule_list *);
+void config_get_access_rule(struct gotwebd_access_rule_list *, struct imsg *);
+void config_free_access_rules(struct gotwebd_access_rule_list *);
+void config_set_repository(struct imsgev *, struct gotwebd_repo *);
+void config_get_repository(struct gotwebd_repolist *, struct imsg *);
+void config_free_repos(struct gotwebd_repolist *);
 int config_init(struct gotwebd *);
