@@ -47,13 +47,13 @@
 #include "got_opentemp.h"
 #include "got_fetch.h"
 
+#include "got_lib_hash.h"
 #include "got_lib_delta.h"
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
 #include "got_lib_object_parse.h"
 #include "got_lib_object_create.h"
 #include "got_lib_pack.h"
-#include "got_lib_hash.h"
 #include "got_lib_privsep.h"
 #include "got_lib_object_cache.h"
 #include "got_lib_repository.h"
@@ -74,7 +74,8 @@
 
 const struct got_error *
 got_fetch_connect(pid_t *fetchpid, int *fetchfd, const char *proto,
-    const char *host, const char *port, const char *server_path, int verbosity)
+    const char *host, const char *port, const char *server_path,
+    const char *jumphost, const char *identity_file, int verbosity)
 {
 	const struct got_error *err = NULL;
 
@@ -83,7 +84,8 @@ got_fetch_connect(pid_t *fetchpid, int *fetchfd, const char *proto,
 
 	if (strcmp(proto, "ssh") == 0 || strcmp(proto, "git+ssh") == 0)
 		err = got_dial_ssh(fetchpid, fetchfd, host, port,
-		    server_path, GOT_DIAL_CMD_FETCH, verbosity);
+		    server_path, jumphost, identity_file, GOT_DIAL_CMD_FETCH,
+		    verbosity);
 	else if (strcmp(proto, "git") == 0)
 		err = got_dial_git(fetchfd, host, port, server_path,
 		    GOT_DIAL_CMD_FETCH);
@@ -120,7 +122,7 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 	char *packpath = NULL, *idxpath = NULL, *id_str = NULL;
 	const char *repo_path = NULL;
 	struct got_pathlist_head have_refs;
-	struct got_pathlist_entry *pe;
+	struct got_pathlist_entry *pe, *new;
 	struct got_reflist_head my_refs;
 	struct got_reflist_entry *re;
 	off_t packfile_size = 0;
@@ -130,12 +132,18 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 	char *progress = NULL;
 
 	*pack_hash = NULL;
+	memset(&fetchibuf, 0, sizeof(fetchibuf));
+	memset(&idxibuf, 0, sizeof(idxibuf));
+
+	if (repo && got_repo_get_object_format(repo) != GOT_HASH_SHA1)
+		return got_error_fmt(GOT_ERR_NOT_IMPL,
+		    "sha256 object IDs unsupported in network protocol");
 
 	/*
 	 * Prevent fetching of references that won't make any
 	 * sense outside of the remote repository's context.
 	 */
-	TAILQ_FOREACH(pe, wanted_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, wanted_refs) {
 		const char *refname = pe->path;
 		if (strncmp(refname, "refs/got/", 9) == 0 ||
 		    strncmp(refname, "got/", 4) == 0 ||
@@ -150,7 +158,7 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 	for (i = 0; i < nitems(tmpfds); i++)
 		tmpfds[i] = -1;
 
-	TAILQ_INIT(&have_refs);
+	RB_INIT(&have_refs);
 	TAILQ_INIT(&my_refs);
 
 	if (!list_refs_only) {
@@ -162,7 +170,7 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 
 	TAILQ_FOREACH(re, &my_refs, entry) {
 		struct got_object_id *id;
-		const char *refname;
+		char *refname;
 
 		if (got_ref_is_symbolic(re->ref))
 			continue;
@@ -175,9 +183,14 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 			err = got_error_from_errno("strdup");
 			goto done;
 		}
-		err = got_pathlist_append(&have_refs, refname, id);
+		err = got_pathlist_insert(&new, &have_refs, refname, id);
 		if (err)
 			goto done;
+		if (new == NULL){
+			free(refname);
+			free(id);
+		}
+
 	}
 
 	if (list_refs_only) {
@@ -254,7 +267,11 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 		err = got_error_from_errno("close");
 		goto done;
 	}
-	imsg_init(&fetchibuf, imsg_fetchfds[0]);
+	if (imsgbuf_init(&fetchibuf, imsg_fetchfds[0]) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		goto done;
+	}
+	imsgbuf_allow_fdpass(&fetchibuf);
 	nfetchfd = dup(fetchfd);
 	if (nfetchfd == -1) {
 		err = got_error_from_errno("dup");
@@ -297,16 +314,20 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 
 		err = got_privsep_recv_fetch_progress(&done,
 		    &id, &refname, symrefs, &server_progress,
-		    &packfile_size_cur, (*pack_hash)->sha1, &fetchibuf);
+		    &packfile_size_cur, (*pack_hash)->hash, &fetchibuf);
 		if (err != NULL)
 			goto done;
 		/* Don't report size progress for an empty pack file. */
 		if (packfile_size_cur <= ssizeof(pack_hdr) + SHA1_DIGEST_LENGTH)
 			packfile_size_cur = 0;
 		if (!done && refname && id) {
-			err = got_pathlist_insert(NULL, refs, refname, id);
-			if (err)
-				goto done;
+			err = got_pathlist_insert(&pe, refs, refname, id);
+			if (err || pe == NULL) {
+				free(id);
+				free(refname);
+				if (err != NULL)
+					goto done;
+			}
 		} else if (!done && server_progress) {
 			char *p;
 			/*
@@ -443,15 +464,18 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 		err = got_error_from_errno("close");
 		goto done;
 	}
-	imsg_init(&idxibuf, imsg_idxfds[0]);
+	if (imsgbuf_init(&idxibuf, imsg_idxfds[0]) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		goto done;
+	}
+	imsgbuf_allow_fdpass(&idxibuf);
 
 	npackfd = dup(packfd);
 	if (npackfd == -1) {
 		err = got_error_from_errno("dup");
 		goto done;
 	}
-	err = got_privsep_send_index_pack_req(&idxibuf, (*pack_hash)->sha1,
-	    npackfd);
+	err = got_privsep_send_index_pack_req(&idxibuf, *pack_hash, npackfd);
 	if (err != NULL)
 		goto done;
 	npackfd = -1;
@@ -522,6 +546,10 @@ got_fetch_pack(struct got_object_id **pack_hash, struct got_pathlist_head *refs,
 	tmpidxpath = NULL;
 
 done:
+	if (fetchibuf.w)
+		imsgbuf_clear(&fetchibuf);
+	if (idxibuf.w)
+		imsgbuf_clear(&idxibuf);
 	if (tmppackpath && unlink(tmppackpath) == -1 && err == NULL)
 		err = got_error_from_errno2("unlink", tmppackpath);
 	if (tmpidxpath && unlink(tmpidxpath) == -1 && err == NULL)
