@@ -24,6 +24,7 @@ struct got_pack_privsep_child {
 struct got_pack {
 	char *path_packfile;
 	int fd;
+	enum got_hash_algorithm algo;
 	uint8_t *map;
 	off_t filesize;
 	struct got_pack_privsep_child *privsep_child;
@@ -56,19 +57,15 @@ const struct got_error *got_pack_parse_object_type_and_size(uint8_t *,
 #define GOT_PACKFILE_NAMELEN	(strlen(GOT_PACK_PREFIX) + \
 				SHA1_DIGEST_STRING_LENGTH - 1 + \
 				strlen(GOT_PACKFILE_SUFFIX))
-#define GOT_PACKIDX_NAMELEN	(strlen(GOT_PACK_PREFIX) + \
-				SHA1_DIGEST_STRING_LENGTH - 1 + \
-				strlen(GOT_PACKIDX_SUFFIX))
+#define GOT_PACKIDX_NAMELEN(digest_len) \
+				(strlen(GOT_PACK_PREFIX) + \
+				digest_len - 1 + strlen(GOT_PACKIDX_SUFFIX))
 
 /* See Documentation/technical/pack-format.txt in Git. */
 
 struct got_packidx_trailer {
-	u_int8_t	packfile_sha1[SHA1_DIGEST_LENGTH];
-	u_int8_t	packidx_sha1[SHA1_DIGEST_LENGTH];
-} __attribute__((__packed__));
-
-struct got_packidx_object_id {
-	u_int8_t sha1[SHA1_DIGEST_LENGTH];
+	u_int8_t	packfile_hash[GOT_HASH_DIGEST_MAXLEN];
+	u_int8_t	packidx_hash[GOT_HASH_DIGEST_MAXLEN];
 } __attribute__((__packed__));
 
 /* Ignore pack index version 1 which is no longer written by Git. */
@@ -81,17 +78,17 @@ struct got_packidx_v2_hdr {
 
 	/*
 	 * Each entry N in the fanout table contains the number of objects in
-	 * the packfile whose SHA1 begins with a byte less than or equal to N.
+	 * the packfile whose hash begins with a byte less than or equal to N.
 	 * The last entry (index 255) contains the number of objects in the
-	 * pack file whose first SHA1 byte is <= 0xff, and thus records the
+	 * pack file whose first hash byte is <= 0xff, and thus records the
 	 * total number of objects in the pack file. All pointer variables
 	 * below point to tables with a corresponding number of entries.
 	 */
 	uint32_t	*fanout_table;	/* values are big endian */
 #define GOT_PACKIDX_V2_FANOUT_TABLE_ITEMS (0xff + 1)
 
-	/* Sorted SHA1 checksums for each object in the pack file. */
-	struct got_packidx_object_id *sorted_ids;
+	/* Sorted checksums for each object in the pack file. */
+	uint8_t *sorted_ids;
 
 	/* CRC32 of the packed representation of each object. */
 	uint32_t	*crc32;
@@ -104,7 +101,7 @@ struct got_packidx_v2_hdr {
 	/* Large offsets table is empty for pack files < 2 GB. */
 	uint64_t	*large_offsets;		/* values are big endian */
 
-	struct got_packidx_trailer *trailer;
+	struct got_packidx_trailer trailer;
 };
 
 struct got_pack_offset_index {
@@ -121,6 +118,7 @@ struct got_pack_large_offset_index {
 struct got_packidx {
 	char *path_packidx; /* actual on-disk path */
 	int fd;
+	enum got_hash_algorithm algo;
 	uint8_t *map;
 	size_t len;
 	size_t nlargeobj;
@@ -159,7 +157,7 @@ struct got_packfile_obj_hdr {
 
 const struct got_error *got_packidx_init_hdr(struct got_packidx *, int, off_t);
 const struct got_error *got_packidx_open(struct got_packidx **,
-    int, const char *, int);
+    int, const char *, int, enum got_hash_algorithm);
 const struct got_error *got_packidx_close(struct got_packidx *);
 const struct got_error *got_packidx_get_packfile_path(char **, const char *);
 off_t got_packidx_get_object_offset(struct got_packidx *, int idx);
