@@ -72,15 +72,15 @@ const struct got_error *
 got_packidx_init_hdr(struct got_packidx *p, int verify, off_t packfile_size)
 {
 	const struct got_error *err = NULL;
-	enum got_hash_algorithm algo = GOT_HASH_SHA1;
 	struct got_packidx_v2_hdr *h;
 	struct got_hash ctx;
 	uint8_t hash[GOT_HASH_DIGEST_MAXLEN];
-	size_t nobj, len_fanout, len_ids, offset, remain;
+	size_t nobj, len_fanout, len_ids, offset, remain, digest_string_len;
 	ssize_t n;
 	int i;
 
-	got_hash_init(&ctx, algo);
+	got_hash_init(&ctx, p->algo);
+	digest_string_len = got_hash_digest_length(p->algo);
 
 	h = &p->hdr;
 	offset = 0;
@@ -180,14 +180,13 @@ got_packidx_init_hdr(struct got_packidx *p, int verify, off_t packfile_size)
 	remain -= len_fanout;
 
 	nobj = be32toh(h->fanout_table[0xff]);
-	len_ids = nobj * sizeof(*h->sorted_ids);
+	len_ids = nobj * got_hash_digest_length(p->algo);
 	if (len_ids <= nobj || len_ids > remain) {
 		err = got_error(GOT_ERR_BAD_PACKIDX);
 		goto done;
 	}
 	if (p->map)
-		h->sorted_ids =
-		    (struct got_packidx_object_id *)((uint8_t*)(p->map + offset));
+		h->sorted_ids = p->map + offset;
 	else {
 		h->sorted_ids = malloc(len_ids);
 		if (h->sorted_ids == NULL) {
@@ -301,32 +300,36 @@ got_packidx_init_hdr(struct got_packidx *p, int verify, off_t packfile_size)
 	offset += p->nlargeobj * sizeof(*h->large_offsets);
 
 checksum:
-	if (remain < sizeof(*h->trailer)) {
+	if (remain < digest_string_len * 2) {
 		err = got_error(GOT_ERR_BAD_PACKIDX);
 		goto done;
 	}
-	if (p->map)
-		h->trailer =
-		    (struct got_packidx_trailer *)((uint8_t*)(p->map + offset));
-	else {
-		h->trailer = malloc(sizeof(*h->trailer));
-		if (h->trailer == NULL) {
-			err = got_error_from_errno("malloc");
-			goto done;
-		}
-		n = read(p->fd, h->trailer, sizeof(*h->trailer));
+	if (p->map) {
+		memcpy(h->trailer.packfile_hash, p->map + offset,
+		    digest_string_len);
+		memcpy(h->trailer.packidx_hash,
+		    p->map + offset + digest_string_len, digest_string_len);
+	} else {
+		n = read(p->fd, h->trailer.packfile_hash, digest_string_len);
 		if (n < 0)
 			err = got_error_from_errno("read");
-		else if (n != sizeof(*h->trailer)) {
+		else if (n != digest_string_len) {
+			err = got_error(GOT_ERR_BAD_PACKIDX);
+			goto done;
+		}
+		n = read(p->fd, h->trailer.packidx_hash, digest_string_len);
+		if (n < 0)
+			err = got_error_from_errno("read");
+		else if (n != digest_string_len) {
 			err = got_error(GOT_ERR_BAD_PACKIDX);
 			goto done;
 		}
 	}
 	if (verify) {
-		got_hash_update(&ctx, h->trailer->packfile_sha1,
-		    SHA1_DIGEST_LENGTH);
+		got_hash_update(&ctx, h->trailer.packfile_hash,
+		    digest_string_len);
 		got_hash_final(&ctx, hash);
-		if (got_hash_cmp(ctx.algo, hash, h->trailer->packidx_sha1) != 0)
+		if (got_hash_cmp(ctx.algo, hash, h->trailer.packidx_hash) != 0)
 			err = got_error(GOT_ERR_PACKIDX_CSUM);
 	}
 done:
@@ -335,7 +338,8 @@ done:
 
 const struct got_error *
 got_packidx_open(struct got_packidx **packidx,
-    int dir_fd, const char *relpath, int verify)
+    int dir_fd, const char *relpath, int verify,
+    enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
 	struct got_packidx *p = NULL;
@@ -354,10 +358,9 @@ got_packidx_open(struct got_packidx **packidx,
 	 * the existence of lonely pack index files but we do not.
 	 */
 	if (fstatat(dir_fd, pack_relpath, &pack_sb, 0) == -1) {
-		if (errno == ENOENT) {
-			err = got_error_fmt(GOT_ERR_LONELY_PACKIDX,
-			    "%s", relpath);
-		} else
+		if (errno == ENOENT)
+			err = got_error_path(relpath, GOT_ERR_LONELY_PACKIDX);
+		else
 			err = got_error_from_errno2("fstatat", pack_relpath);
 		goto done;
 	}
@@ -368,6 +371,8 @@ got_packidx_open(struct got_packidx **packidx,
 		goto done;
 	}
 
+	p->algo = algo;
+
 	p->fd = openat(dir_fd, relpath, O_RDONLY | O_NOFOLLOW | O_CLOEXEC);
 	if (p->fd == -1) {
 		err = got_error_from_errno2("openat", relpath);
@@ -431,7 +436,6 @@ got_packidx_close(struct got_packidx *packidx)
 		free(packidx->hdr.crc32);
 		free(packidx->hdr.offsets);
 		free(packidx->hdr.large_offsets);
-		free(packidx->hdr.trailer);
 	}
 	if (close(packidx->fd) == -1 && err == NULL)
 		err = got_error_from_errno("close");
@@ -487,20 +491,21 @@ int
 got_packidx_get_object_idx(struct got_packidx *packidx,
     struct got_object_id *id)
 {
-	u_int8_t id0 = id->sha1[0];
+	u_int8_t id0 = id->hash[0];
 	uint32_t totobj = be32toh(packidx->hdr.fanout_table[0xff]);
 	int left = 0, right = totobj - 1;
+	size_t digest_len = got_hash_digest_length(packidx->algo);
 
 	if (id0 > 0)
 		left = be32toh(packidx->hdr.fanout_table[id0 - 1]);
 
 	while (left <= right) {
-		struct got_packidx_object_id *oid;
+		uint8_t *oid;
 		int i, cmp;
 
 		i = ((left + right) / 2);
-		oid = &packidx->hdr.sorted_ids[i];
-		cmp = memcmp(id->sha1, oid->sha1, SHA1_DIGEST_LENGTH);
+		oid = packidx->hdr.sorted_ids + i * digest_len;
+		cmp = memcmp(id->hash, oid, digest_len);
 		if (cmp == 0)
 			return i;
 		else if (cmp > 0)
@@ -650,13 +655,15 @@ got_packidx_get_object_id(struct got_object_id *id,
     struct got_packidx *packidx, int idx)
 {
 	uint32_t totobj = be32toh(packidx->hdr.fanout_table[0xff]);
-	struct got_packidx_object_id *oid;
+	uint8_t *oid;
+	size_t digest_len = got_hash_digest_length(packidx->algo);
 
 	if (idx < 0 || idx >= totobj)
 		return got_error(GOT_ERR_NO_OBJ);
 
-	oid = &packidx->hdr.sorted_ids[idx];
-	memcpy(id->sha1, oid->sha1, SHA1_DIGEST_LENGTH);
+	oid = packidx->hdr.sorted_ids + idx * digest_len;
+	memcpy(id->hash, oid, digest_len);
+	id->algo = packidx->algo;
 	return NULL;
 }
 
@@ -669,8 +676,9 @@ got_packidx_match_id_str_prefix(struct got_object_id_queue *matched_ids,
 	uint32_t totobj = be32toh(packidx->hdr.fanout_table[0xff]);
 	char hex[3];
 	size_t prefix_len = strlen(id_str_prefix);
-	struct got_packidx_object_id *oid;
+	uint8_t *oid;
 	uint32_t i = 0;
+	size_t digest_len = got_hash_digest_length(packidx->algo);
 
 	if (prefix_len < 2)
 		return got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
@@ -683,18 +691,19 @@ got_packidx_match_id_str_prefix(struct got_object_id_queue *matched_ids,
 
 	if (id0 > 0)
 		i = be32toh(packidx->hdr.fanout_table[id0 - 1]);
-	oid = &packidx->hdr.sorted_ids[i];
-	while (i < totobj && oid->sha1[0] == id0) {
-		char id_str[SHA1_DIGEST_STRING_LENGTH];
+	oid = packidx->hdr.sorted_ids + i * digest_len;
+	while (i < totobj && oid[0] == id0) {
+		char id_str[GOT_HASH_DIGEST_STRING_MAXLEN];
 		struct got_object_qid *qid;
 		int cmp;
 
-		if (!got_sha1_digest_to_str(oid->sha1, id_str, sizeof(id_str)))
+		if (!got_hash_digest_to_str(oid, id_str, sizeof(id_str),
+		    packidx->algo))
 			return got_error(GOT_ERR_NO_SPACE);
 
 		cmp = strncmp(id_str, id_str_prefix, prefix_len);
 		if (cmp < 0) {
-			oid = &packidx->hdr.sorted_ids[++i];
+			oid = packidx->hdr.sorted_ids + (++i) * digest_len;
 			continue;
 		} else if (cmp > 0)
 			break;
@@ -702,10 +711,11 @@ got_packidx_match_id_str_prefix(struct got_object_id_queue *matched_ids,
 		err = got_object_qid_alloc_partial(&qid);
 		if (err)
 			return err;
-		memcpy(qid->id.sha1, oid->sha1, SHA1_DIGEST_LENGTH);
+		memcpy(qid->id.hash, oid, digest_len);
+		qid->id.algo = packidx->algo;
 		STAILQ_INSERT_TAIL(matched_ids, qid, entry);
 
-		oid = &packidx->hdr.sorted_ids[++i];
+		oid = packidx->hdr.sorted_ids + (++i) * digest_len;
 	}
 
 	return NULL;
@@ -752,6 +762,8 @@ got_pack_start_privsep_child(struct got_pack *pack, struct got_packidx *packidx)
 	pid = fork();
 	if (pid == -1) {
 		err = got_error_from_errno("fork");
+		close(imsg_fds[0]);
+		close(imsg_fds[1]);
 		goto done;
 	} else if (pid == 0) {
 		set_max_datasize();
@@ -760,11 +772,20 @@ got_pack_start_privsep_child(struct got_pack *pack, struct got_packidx *packidx)
 		/* not reached */
 	}
 
-	if (close(imsg_fds[1]) == -1)
-		return got_error_from_errno("close");
+	if (close(imsg_fds[1]) == -1) {
+		err = got_error_from_errno("close");
+		close(imsg_fds[0]);
+		goto done;
+	}
 	pack->privsep_child->imsg_fd = imsg_fds[0];
 	pack->privsep_child->pid = pid;
-	imsg_init(ibuf, imsg_fds[0]);
+	if (imsgbuf_init(ibuf, imsg_fds[0]) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		close(imsg_fds[0]);
+		goto done;
+	}
+	imsgbuf_allow_fdpass(ibuf);
+
 	pack->privsep_child->ibuf = ibuf;
 
 	err = got_privsep_init_pack_child(ibuf, pack, packidx);
@@ -778,6 +799,7 @@ got_pack_start_privsep_child(struct got_pack *pack, struct got_packidx *packidx)
 	}
 done:
 	if (err) {
+		imsgbuf_clear(ibuf);
 		free(ibuf);
 		free(pack->privsep_child);
 		pack->privsep_child = NULL;
@@ -788,20 +810,21 @@ done:
 static const struct got_error *
 pack_stop_privsep_child(struct got_pack *pack)
 {
-	const struct got_error *err = NULL, *close_err = NULL;
+	const struct got_error *err = NULL;
+	const struct got_error *close_err = NULL, *child_err = NULL;
 
 	if (pack->privsep_child == NULL)
 		return NULL;
 
 	err = got_privsep_send_stop(pack->privsep_child->imsg_fd);
-	if (err)
-		return err;
 	if (close(pack->privsep_child->imsg_fd) == -1)
 		close_err = got_error_from_errno("close");
-	err = got_privsep_wait_for_child(pack->privsep_child->pid);
 	if (close_err && err == NULL)
 		err = close_err;
-	imsg_clear(pack->privsep_child->ibuf);
+	child_err = got_privsep_wait_for_child(pack->privsep_child->pid);
+	if (child_err && err == NULL)
+		err = child_err;
+	imsgbuf_clear(pack->privsep_child->ibuf);
 	free(pack->privsep_child->ibuf);
 	free(pack->privsep_child);
 	pack->privsep_child = NULL;
@@ -1092,6 +1115,11 @@ const struct got_error *
 got_pack_parse_ref_delta(struct got_object_id *id,
     struct got_pack *pack, off_t delta_offset, int tslen)
 {
+	size_t digest_len = got_hash_digest_length(pack->algo);
+
+	memset(id, 0, sizeof(*id));
+	id->algo = pack->algo;
+
 	if (pack->map) {
 		size_t mapoff;
 
@@ -1104,13 +1132,13 @@ got_pack_parse_ref_delta(struct got_object_id *id,
 		mapoff = delta_offset + tslen;
 		if (mapoff + sizeof(*id) >= pack->filesize)
 			return got_error(GOT_ERR_PACK_OFFSET);
-		memcpy(id, pack->map + mapoff, sizeof(*id));
+		memcpy(id->hash, pack->map + mapoff, digest_len);
 	} else {
 		ssize_t n;
-		n = read(pack->fd, id, sizeof(*id));
+		n = read(pack->fd, id->hash, digest_len);
 		if (n < 0)
 			return got_error_from_errno("read");
-		if (n != sizeof(*id))
+		if (n != digest_len)
 			return got_error(GOT_ERR_BAD_PACKFILE);
 	}
 
@@ -1138,7 +1166,8 @@ resolve_ref_delta(struct got_delta_chain *deltas, struct got_packidx *packidx,
 	if (err)
 		return err;
 	if (pack->map) {
-		delta_data_offset = delta_offset + tslen + SHA1_DIGEST_LENGTH;
+		delta_data_offset = delta_offset + tslen +
+		    got_hash_digest_length(packidx->algo);
 	} else {
 		delta_data_offset = lseek(pack->fd, 0, SEEK_CUR);
 		if (delta_data_offset == -1)
@@ -2081,7 +2110,7 @@ got_packfile_extract_raw_delta(uint8_t **delta_buf, size_t *delta_size,
 		err = got_pack_parse_ref_delta(base_id, pack, offset, tslen);
 		if (err)
 			return err;
-		delta_hdrlen = SHA1_DIGEST_LENGTH;
+		delta_hdrlen = got_hash_digest_length(pack->algo);
 		break;
 	default:
 		return got_error_fmt(GOT_ERR_OBJ_TYPE,
