@@ -93,38 +93,6 @@ send_idset(struct imsgbuf *ibuf, struct got_object_idset *idset)
 	return got_privsep_send_object_idlist_done(ibuf);
 }
 
-static const struct got_error *
-send_filtered_id_queue(struct imsgbuf *ibuf, struct got_object_id_queue *ids,
-    uintptr_t color)
-{
-	const struct got_error *err = NULL;
-	struct got_object_qid *qid;
-	struct got_object_id *filtered_ids[GOT_IMSG_OBJ_ID_LIST_MAX_NIDS];
-	int nids = 0;
-
-	STAILQ_FOREACH(qid, ids, entry) {
-		if (color != (intptr_t)qid->data)
-			continue;
-
-		filtered_ids[nids++] = &qid->id;
-		if (nids >= GOT_IMSG_OBJ_ID_LIST_MAX_NIDS) {
-			err = got_privsep_send_object_idlist(ibuf,
-			    filtered_ids, nids);
-			if (err)
-				return err;
-			nids = 0;
-		}
-	}
-
-	if (nids > 0) {
-		err = got_privsep_send_object_idlist(ibuf, filtered_ids, nids);
-		if (err)
-			return err;
-	}
-
-	return got_privsep_send_object_idlist_done(ibuf);
-}
-
 static const struct got_error *
 recv_reused_delta(struct got_imsg_reused_delta *delta,
     struct got_object_idset *idset, struct got_pack_metavec *v)
@@ -249,7 +217,7 @@ recv_painted_commit(void *arg, struct got_object_id *id, intptr_t color)
 {
 	const struct got_error *err = NULL;
 	struct recv_painted_commit_arg *a = arg;
-	struct got_object_qid *qid, *tmp;
+	struct got_object_qid *qid;
 
 	if (a->cancel_cb) {
 		err = a->cancel_cb(a->cancel_arg);
@@ -286,17 +254,18 @@ recv_painted_commit(void *arg, struct got_object_id *id, intptr_t color)
 		    "%s invalid commit color %"PRIdPTR, __func__, color);
 	}
 
-	STAILQ_FOREACH_SAFE(qid, a->ids, entry, tmp) {
+	STAILQ_FOREACH(qid, a->ids, entry) {
+		int ocolor;
 		if (got_object_id_cmp(&qid->id, id) != 0)
 			continue;
-		STAILQ_REMOVE(a->ids, qid, got_object_qid, entry);
-		color = (intptr_t)qid->data;
-		if (*(a->qid0) == qid)
-			*(a->qid0) = NULL;
-		got_object_qid_free(qid);
-		(*a->nqueued)--;
-		if (color == COLOR_SKIP)
-			(*a->nskip)--;
+		ocolor = (intptr_t)qid->data;
+		if (ocolor != color) {
+			got_pack_paint_commit(qid, color);
+			if (ocolor == COLOR_SKIP)
+				(*a->nskip)--;
+			else if (color == COLOR_SKIP)
+				(*a->nskip)++;
+		}
 		break;
 	}
 
@@ -324,18 +293,8 @@ paint_packed_commits(struct got_object_qid **qid0,
 	if (err)
 		return err;
 
-	err = send_filtered_id_queue(pack->privsep_child->ibuf, ids,
-	    COLOR_KEEP);
-	if (err)
-		return err;
-
-	err = send_filtered_id_queue(pack->privsep_child->ibuf, ids,
-	    COLOR_DROP);
-	if (err)
-		return err;
-
-	err = send_filtered_id_queue(pack->privsep_child->ibuf, ids,
-	    COLOR_SKIP);
+	err = got_privsep_send_object_id_queue(pack->privsep_child->ibuf,
+	    ids, *nqueued);
 	if (err)
 		return err;
 
@@ -382,7 +341,10 @@ paint_packed_commits(struct got_object_qid **qid0,
 		qid = STAILQ_FIRST(&next_ids);
 		STAILQ_REMOVE_HEAD(&next_ids, entry);
 		got_pack_paint_commit(qid, color);
-		STAILQ_INSERT_TAIL(ids, qid, entry);
+		if (color == COLOR_KEEP)
+			STAILQ_INSERT_TAIL(ids, qid, entry);
+		else
+			STAILQ_INSERT_HEAD(ids, qid, entry);
 		(*nqueued)++;
 		if (color == COLOR_SKIP)
 			(*nskip)++;
@@ -483,17 +445,15 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				    ids, keep, drop, skip, repo,
 				    progress_cb, progress_arg, rl,
 				    cancel_cb, cancel_arg);
-				if (err && qid == NULL)
-					goto done;
 				if (qid) {
 					STAILQ_REMOVE(ids, qid,
 					    got_object_qid, entry);
 					nqueued--;
 					got_object_qid_free(qid);
 					qid = NULL;
-					if (err)
-						goto done;
 				}
+				if (err)
+					goto done;
 				continue;
 			}
 		}
@@ -536,10 +496,6 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				    NULL);
 				if (err)
 					goto done;
-				err = got_pack_repaint_parent_commits(&qid->id,
-				    COLOR_SKIP, skip, skip, repo);
-				if (err)
-					goto done;
 			} else
 				(*ncolored)++;
 			err = got_object_idset_add(keep, &qid->id, NULL);
@@ -555,10 +511,6 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				    NULL);
 				if (err)
 					goto done;
-				err = got_pack_repaint_parent_commits(&qid->id,
-				    COLOR_SKIP, skip, skip, repo);
-				if (err)
-					goto done;
 			} else
 				(*ncolored)++;
 			err = got_object_idset_add(drop, &qid->id, NULL);
