@@ -156,11 +156,14 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 	const char *obj_label;
 	size_t mapoff = obj->off;
 	struct got_inflate_checksum csum;
+	size_t digest_len;
 
 	memset(&csum, 0, sizeof(csum));
 	csum.input_ctx = pack_hash_ctx;
 	csum.input_crc = &obj->crc;
 
+	digest_len = got_hash_digest_length(pack->algo);
+
 	err = got_pack_parse_object_type_and_size(&obj->type, &obj->size,
 	    &obj->tslen, pack, obj->off);
 	if (err)
@@ -210,7 +213,7 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 		}
 		if (err)
 			break;
-		got_hash_init(&ctx, GOT_HASH_SHA1);
+		got_hash_init(&ctx, pack->algo);
 		err = got_object_type_label(&obj_label, obj->type);
 		if (err) {
 			free(data);
@@ -238,48 +241,50 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 		free(data);
 		break;
 	case GOT_OBJ_TYPE_REF_DELTA:
-		memset(obj->id.sha1, 0xff, SHA1_DIGEST_LENGTH);
+		memset(obj->id.hash, 0xff, digest_len);
+		obj->id.algo = pack->algo;
 		if (pack->map) {
-			if (mapoff + SHA1_DIGEST_LENGTH >= pack->filesize) {
+			if (mapoff + digest_len >= pack->filesize) {
 				err = got_error(GOT_ERR_BAD_PACKFILE);
 				break;
 			}
-			memcpy(obj->delta.ref.ref_id.sha1, pack->map + mapoff,
-			    SHA1_DIGEST_LENGTH);
+			memcpy(obj->delta.ref.ref_id.hash, pack->map + mapoff,
+			    digest_len);
 			obj->crc = crc32(obj->crc, pack->map + mapoff,
-			    SHA1_DIGEST_LENGTH);
+			    digest_len);
 			got_hash_update(pack_hash_ctx, pack->map + mapoff,
-			    SHA1_DIGEST_LENGTH);
-			mapoff += SHA1_DIGEST_LENGTH;
+			    digest_len);
+			mapoff += digest_len;
 			err = got_inflate_to_mem_mmap(NULL, &datalen,
 			    &obj->len, &csum, pack->map, mapoff,
 			    pack->filesize - mapoff);
 			if (err)
 				break;
 		} else {
-			n = read(pack->fd, obj->delta.ref.ref_id.sha1,
-			    SHA1_DIGEST_LENGTH);
+			n = read(pack->fd, obj->delta.ref.ref_id.hash,
+			    digest_len);
 			if (n == -1) {
 				err = got_error_from_errno("read");
 				break;
 			}
-			if (n < sizeof(obj->id)) {
+			if (n < digest_len) {
 				err = got_error(GOT_ERR_BAD_PACKFILE);
 				break;
 			}
-			obj->crc = crc32(obj->crc, obj->delta.ref.ref_id.sha1,
-			    SHA1_DIGEST_LENGTH);
+			obj->crc = crc32(obj->crc, obj->delta.ref.ref_id.hash,
+			    digest_len);
 			got_hash_update(pack_hash_ctx,
-			    obj->delta.ref.ref_id.sha1, SHA1_DIGEST_LENGTH);
+			    obj->delta.ref.ref_id.hash, digest_len);
 			err = got_inflate_to_mem_fd(NULL, &datalen, &obj->len,
 			    &csum, obj->size, pack->fd);
 			if (err)
 				break;
 		}
-		obj->len += SHA1_DIGEST_LENGTH;
+		obj->len += digest_len;
 		break;
 	case GOT_OBJ_TYPE_OFFSET_DELTA:
-		memset(obj->id.sha1, 0xff, SHA1_DIGEST_LENGTH);
+		memset(obj->id.hash, 0xff, digest_len);
+		obj->id.algo = pack->algo;
 		err = got_pack_parse_offset_delta(&obj->delta.ofs.base_offset,
 		    &obj->delta.ofs.base_offsetlen, pack, obj->off,
 		    obj->tslen);
@@ -314,8 +319,8 @@ read_packed_object(struct got_pack *pack, struct got_indexed_object *obj,
 				break;
 		} else {
 			/*
-			 * XXX Seek back and get CRC and SHA1 of on-disk
-			 * offset bytes.
+			 * XXX Seek back and get CRC and hash digest
+			 * of on-disk offset bytes.
 			 */
 			if (lseek(pack->fd, obj->off + obj->tslen, SEEK_SET)
 			    == -1) {
@@ -413,7 +418,7 @@ resolve_deltified_object(struct got_pack *pack, struct got_packidx *packidx,
 		goto done;
 	}
 	headerlen = strlen(header) + 1;
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	got_hash_init(&ctx, pack->algo);
 	got_hash_update(&ctx, header, headerlen);
 	if (max_size > GOT_DELTA_RESULT_SIZE_CACHED_MAX) {
 		err = read_file_digest(&ctx, tmpfile, len);
@@ -435,23 +440,24 @@ done:
 
 /* Determine the slot in the pack index a given object ID should use. */
 static int
-find_object_idx(struct got_packidx *packidx, uint8_t *sha1)
+find_object_idx(struct got_packidx *packidx, uint8_t *hash)
 {
-	u_int8_t id0 = sha1[0];
+	u_int8_t id0 = hash[0];
 	uint32_t nindexed = be32toh(packidx->hdr.fanout_table[0xff]);
 	int left = 0, right = nindexed - 1;
 	int cmp = 0, i = 0;
+	size_t digest_len = got_hash_digest_length(packidx->algo);
 
 	if (id0 > 0)
 		left = be32toh(packidx->hdr.fanout_table[id0 - 1]);
 
 	while (left <= right) {
-		struct got_packidx_object_id *oid;
+		uint8_t *oid;
 
 		i = ((left + right) / 2);
-		oid = &packidx->hdr.sorted_ids[i];
+		oid = packidx->hdr.sorted_ids + i * digest_len;
 
-		cmp = memcmp(sha1, oid->sha1, SHA1_DIGEST_LENGTH);
+		cmp = memcmp(hash, oid, digest_len);
 		if (cmp == 0)
 			return -1; /* object already indexed */
 		else if (cmp > 0)
@@ -468,13 +474,14 @@ static void
 print_packidx(struct got_packidx *packidx)
 {
 	uint32_t nindexed = be32toh(packidx->hdr.fanout_table[0xff]);
+	size_t digest_len = got_hash_digest_length(packidx->algo);
 	int i;
 
 	fprintf(stderr, "object IDs:\n");
 	for (i = 0; i < nindexed; i++) {
-		char hex[SHA1_DIGEST_STRING_LENGTH];
-		got_sha1_digest_to_str(packidx->hdr.sorted_ids[i].sha1,
-		    hex, sizeof(hex));
+		char hex[GOT_HASH_DIGEST_STRING_MAXLEN];
+		got_hash_digest_to_str(packidx->hdr.sorted_ids + i * digest_len,
+		    hex, sizeof(hex), packidx->algo);
 		fprintf(stderr, "%s\n", hex);
 	}
 	fprintf(stderr, "\n");
@@ -503,9 +510,12 @@ add_indexed_object(struct got_packidx *packidx, uint32_t idx,
     struct got_indexed_object *obj)
 {
 	int i;
+	uint8_t *oid;
+	size_t digest_len = got_hash_digest_length(packidx->algo);
 
-	memcpy(packidx->hdr.sorted_ids[idx].sha1, obj->id.sha1,
-	    SHA1_DIGEST_LENGTH);
+	oid = packidx->hdr.sorted_ids + idx * digest_len;
+	memcpy(oid, obj->id.hash, digest_len);
+	obj->id.algo = packidx->algo;
 	packidx->hdr.crc32[idx] = htobe32(obj->crc);
 	if (obj->off < GOT_PACKIDX_OFFSET_VAL_IS_LARGE_IDX)
 		packidx->hdr.offsets[idx] = htobe32(obj->off);
@@ -517,7 +527,7 @@ add_indexed_object(struct got_packidx *packidx, uint32_t idx,
 		packidx->nlargeobj++;
 	}
 
-	for (i = obj->id.sha1[0]; i <= 0xff; i++) {
+	for (i = obj->id.hash[0]; i <= 0xff; i++) {
 		uint32_t n = be32toh(packidx->hdr.fanout_table[i]);
 		packidx->hdr.fanout_table[i] = htobe32(n + 1);
 	}
@@ -561,14 +571,16 @@ update_packidx(struct got_packidx *packidx, uint32_t nobj,
 {
 	int idx;
 	uint32_t nindexed = be32toh(packidx->hdr.fanout_table[0xff]);
+	size_t digest_len = got_hash_digest_length(packidx->algo);
+	uint8_t *from, *to;
 
-	idx = find_object_idx(packidx, obj->id.sha1);
+	idx = find_object_idx(packidx, obj->id.hash);
 	if (idx == -1)
 		return; /* object already indexed */
 
-	memmove(&packidx->hdr.sorted_ids[idx + 1],
-	    &packidx->hdr.sorted_ids[idx],
-	    sizeof(struct got_packidx_object_id) * (nindexed - idx));
+	from = packidx->hdr.sorted_ids + idx * digest_len;
+	to = from + digest_len;
+	memmove(to, from, digest_len * (nindexed - idx));
 	memmove(&packidx->hdr.offsets[idx + 1], &packidx->hdr.offsets[idx],
 	    sizeof(uint32_t) * (nindexed - idx));
 
@@ -595,7 +607,8 @@ report_progress(uint32_t nobj_total, uint32_t nobj_indexed, uint32_t nobj_loose,
 
 const struct got_error *
 got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
-    FILE *delta_base_file, FILE *delta_accum_file, uint8_t *pack_sha1_expected,
+    FILE *delta_base_file, FILE *delta_accum_file,
+    struct got_object_id *pack_hash_expected,
     got_pack_index_progress_cb progress_cb, void *progress_arg,
     struct got_ratelimit *rl)
 {
@@ -603,19 +616,23 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 	struct got_packfile_hdr hdr;
 	struct got_packidx packidx;
 	char buf[8];
-	char pack_sha1[SHA1_DIGEST_LENGTH];
+	struct got_object_id pack_hash;
 	uint32_t nobj, nvalid, nloose, nresolved = 0, i;
 	struct got_indexed_object *objects = NULL, *obj;
 	struct got_hash ctx;
-	uint8_t packidx_hash[SHA1_DIGEST_LENGTH];
+	uint8_t packidx_hash[GOT_HASH_DIGEST_MAXLEN];
 	ssize_t r, w;
 	int pass, have_ref_deltas = 0, first_delta_idx = -1;
 	size_t mapoff = 0;
 	int p_indexed = 0, last_p_indexed = -1;
 	int p_resolved = 0, last_p_resolved = -1;
+	ssize_t digest_len;
 
-	/* Require that pack file header and SHA1 trailer are present. */
-	if (pack->filesize < sizeof(hdr) + SHA1_DIGEST_LENGTH)
+	/* This has to be signed for lseek(2) later */
+	digest_len = got_hash_digest_length(pack->algo);
+
+	/* Require that pack file header and hash trailer are present. */
+	if (pack->filesize < sizeof(hdr) + digest_len)
 		return got_error_msg(GOT_ERR_BAD_PACKFILE,
 		    "short pack file");
 
@@ -642,8 +659,8 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 		return got_error_msg(GOT_ERR_BAD_PACKFILE,
 		    "bad packfile with zero objects");
 
-	/* We compute the SHA1 of pack file contents and verify later on. */
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	/* We compute the hash of pack file contents and verify later on. */
+	got_hash_init(&ctx, pack->algo);
 	got_hash_update(&ctx, &hdr, sizeof(hdr));
 
 	/*
@@ -669,8 +686,7 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 		err = got_error_from_errno("calloc");
 		goto done;
 	}
-	packidx.hdr.sorted_ids = calloc(nobj,
-	    sizeof(struct got_packidx_object_id));
+	packidx.hdr.sorted_ids = calloc(nobj, digest_len);
 	if (packidx.hdr.sorted_ids == NULL) {
 		err = got_error_from_errno("calloc");
 		goto done;
@@ -685,6 +701,7 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 		err = got_error_from_errno("calloc");
 		goto done;
 	}
+	packidx.algo = pack->algo;
 	/* Large offsets table is empty for pack files < 2 GB. */
 	if (pack->filesize >= GOT_PACKIDX_OFFSET_VAL_IS_LARGE_IDX) {
 		packidx.hdr.large_offsets = calloc(nobj, sizeof(uint64_t));
@@ -767,14 +784,14 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 	 * Having done a full pass over the pack file and can now
 	 * verify its checksum.
 	 */
-	got_hash_final(&ctx, pack_sha1);
+	got_hash_final_object_id(&ctx, &pack_hash);
 
-	if (memcmp(pack_sha1_expected, pack_sha1, SHA1_DIGEST_LENGTH) != 0) {
+	if (got_object_id_cmp(pack_hash_expected, &pack_hash) != 0) {
 		err = got_error(GOT_ERR_PACKFILE_CSUM);
 		goto done;
 	}
 
-	/* Verify the SHA1 checksum stored at the end of the pack file. */
+	/* Verify the hash checksum stored at the end of the pack file. */
 	if (pack->map) {
 		if (pack->filesize > SIZE_MAX) {
 			err = got_error_fmt(GOT_ERR_RANGE,
@@ -783,26 +800,26 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 			goto done;
 		}
 
-		memcpy(pack_sha1_expected, pack->map +
-		    pack->filesize - SHA1_DIGEST_LENGTH,
-		    SHA1_DIGEST_LENGTH);
+		memcpy(pack_hash_expected, pack->map +
+		    pack->filesize - digest_len,
+		    digest_len);
 	} else {
 		ssize_t n;
-		if (lseek(pack->fd, -SHA1_DIGEST_LENGTH, SEEK_END) == -1) {
+		if (lseek(pack->fd, -digest_len, SEEK_END) == -1) {
 			err = got_error_from_errno("lseek");
 			goto done;
 		}
-		n = read(pack->fd, pack_sha1_expected, SHA1_DIGEST_LENGTH);
+		n = read(pack->fd, pack_hash_expected, digest_len);
 		if (n == -1) {
 			err = got_error_from_errno("read");
 			goto done;
 		}
-		if (n != SHA1_DIGEST_LENGTH) {
+		if (n != digest_len) {
 			err = got_error(GOT_ERR_IO);
 			goto done;
 		}
 	}
-	if (memcmp(pack_sha1, pack_sha1_expected, SHA1_DIGEST_LENGTH) != 0) {
+	if (got_object_id_cmp(pack_hash_expected, &pack_hash) != 0) {
 		err = got_error_msg(GOT_ERR_BAD_PACKFILE,
 		    "bad checksum in pack file trailer");
 		goto done;
@@ -902,7 +919,7 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 	free(objects);
 	objects = NULL;
 
-	got_hash_init(&ctx, GOT_HASH_SHA1);
+	got_hash_init(&ctx, pack->algo);
 	putbe32(buf, GOT_PACKIDX_V2_MAGIC);
 	putbe32(buf + 4, GOT_PACKIDX_VERSION);
 	err = got_pack_hwrite(idxfd, buf, 8, &ctx);
@@ -913,7 +930,7 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 	if (err)
 		goto done;
 	err = got_pack_hwrite(idxfd, packidx.hdr.sorted_ids,
-	    nobj * SHA1_DIGEST_LENGTH, &ctx);
+	    nobj * digest_len, &ctx);
 	if (err)
 		goto done;
 	err = got_pack_hwrite(idxfd, packidx.hdr.crc32,
@@ -930,17 +947,17 @@ got_pack_index(struct got_pack *pack, int idxfd, FILE *tmpfile,
 		if (err)
 			goto done;
 	}
-	err = got_pack_hwrite(idxfd, pack_sha1, SHA1_DIGEST_LENGTH, &ctx);
+	err = got_pack_hwrite(idxfd, &pack_hash.hash, digest_len, &ctx);
 	if (err)
 		goto done;
 
 	got_hash_final(&ctx, packidx_hash);
-	w = write(idxfd, packidx_hash, sizeof(packidx_hash));
+	w = write(idxfd, packidx_hash, digest_len);
 	if (w == -1) {
 		err = got_error_from_errno("write");
 		goto done;
 	}
-	if (w != sizeof(packidx_hash)) {
+	if (w != digest_len) {
 		err = got_error(GOT_ERR_IO);
 		goto done;
 	}
