@@ -349,6 +349,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 	char *id_str = NULL, *default_id_str = NULL, *refname = NULL;
 	char *server_capabilities = NULL, *my_capabilities = NULL;
 	const char *default_branch = NULL;
+	char *free_default_branch = NULL;
 	struct got_pathlist_head symrefs;
 	struct got_pathlist_entry *pe;
 	int sent_my_capabilites = 0, have_sidebands = 0;
@@ -444,6 +445,14 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 			}
 			continue;
 		}
+		if (default_branch == NULL &&
+		    strcmp(refname, GOT_REF_HEAD) == 0) {
+			default_id_str = strdup(id_str);
+			if (default_id_str == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
+		}
 		if (default_branch && default_id_str == NULL &&
 		    strcmp(refname, default_branch) == 0) {
 			default_id_str = strdup(id_str);
@@ -452,6 +461,17 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 				goto done;
 			}
 		}
+		/* If no symrefs is given, go for first matching id_str */
+		if (default_branch == NULL && default_id_str &&
+		    strncmp(refname, "refs/heads/", 11) == 0 &&
+		    strcmp(id_str, default_id_str) == 0) {
+			free_default_branch = strdup(refname);
+			if (free_default_branch == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
+			default_branch = free_default_branch;
+		}
 
 		if (list_refs_only || strncmp(refname, "refs/tags/", 10) == 0) {
 			err = fetch_ref(ibuf, have_refs, &have[nref],
@@ -481,7 +501,8 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 					goto done;
 				nref++;
 				found_branch = 1;
-			} else if (chattygot) {
+			} else if (chattygot && (default_branch == NULL ||
+			    strcmp(default_branch, refname) != 0)) {
 				fprintf(stderr, "%s: ignoring %s\n",
 				    getprogname(), refname);
 			}
@@ -844,6 +865,7 @@ done:
 	free(have);
 	free(want);
 	free(id_str);
+	free(free_default_branch);
 	free(default_id_str);
 	free(refname);
 	free(server_capabilities);
