@@ -1673,127 +1673,13 @@ queue_commit_id(struct got_object_id_queue *ids, struct got_object_id *id,
 		return err;
 
 	memcpy(&qid->id, id, sizeof(qid->id));
-	STAILQ_INSERT_TAIL(ids, qid, entry);
+	if (color == COLOR_KEEP)
+		STAILQ_INSERT_TAIL(ids, qid, entry);
+	else
+		STAILQ_INSERT_HEAD(ids, qid, entry);
 	return paint_commit(qid, color);
 }
 
-static const struct got_error *
-repaint_parent_commits(struct got_object_id *commit_id, int commit_idx,
-    int color, struct got_object_idset *set, struct got_object_idset *skip,
-    struct got_object_id_queue *ids, int *nids,
-    struct got_object_id_queue *painted, int *npainted,
-    struct got_pack *pack, struct got_packidx *packidx,
-    struct got_object_cache *objcache)
-{
-	const struct got_error *err;
-	const struct got_object_id_queue *parents;
-	struct got_commit_object *commit;
-	struct got_object_id_queue repaint;
-
-	STAILQ_INIT(&repaint);
-
-	err = open_commit(&commit, pack, packidx, commit_idx, commit_id,
-	    objcache);
-	if (err)
-		return err;
-
-	while (commit) {
-		struct got_object_qid *pid, *qid;
-		int idx;
-
-		parents = got_object_commit_get_parent_ids(commit);
-		if (parents) {
-			STAILQ_FOREACH(pid, parents, entry) {
-				idx = got_packidx_get_object_idx(packidx,
-				    &pid->id);
-				/*
-				 * No need to traverse parents which are not in
-				 * the pack file, are already in the desired
-				 * set, or are marked for skipping already.
-				 */
-				if (idx == -1)
-					continue;
-				if (got_object_idset_contains(set, &pid->id))
-					continue;
-				if (set != skip &&
-				    got_object_idset_contains(skip, &pid->id))
-					continue;
-
-				err = queue_commit_id(&repaint, &pid->id,
-				    color);
-				if (err)
-					goto done;
-			}
-		}
-		got_object_commit_close(commit);
-		commit = NULL;
-
-		pid = STAILQ_FIRST(&repaint);
-		if (pid == NULL)
-			break;
-
-		err = paint_commit(pid, color);
-		if (err)
-			break;
-
-		err = got_object_idset_add(set, &pid->id, NULL);
-		if (err)
-			break;
-
-		STAILQ_REMOVE_HEAD(&repaint, entry);
-
-		/* Insert or replace this commit on the painted list. */
-		STAILQ_FOREACH(qid, painted, entry) {
-			if (got_object_id_cmp(&qid->id, &pid->id) != 0)
-				continue;
-			err = paint_commit(qid, color);
-			if (err)
-				goto done;
-			got_object_qid_free(pid);
-			pid = qid;
-			break;
-		}
-		if (qid == NULL) {
-			STAILQ_INSERT_TAIL(painted, pid, entry);
-			(*npainted)++;
-		}
-
-		/*
-		 * In case this commit is on the caller's list of
-		 * pending commits to traverse, repaint it there.
-		 */
-		STAILQ_FOREACH(qid, ids, entry) {
-			if (got_object_id_cmp(&qid->id, &pid->id) != 0)
-				continue;
-			err = paint_commit(qid, color);
-			if (err)
-				goto done;
-			break;
-		}
-
-		idx = got_packidx_get_object_idx(packidx, &pid->id);
-		if (idx == -1) {
-			/*
-			 * Should not happen because we only queue
-			 * parents which exist in our pack file.
-			 */
-			err = got_error(GOT_ERR_NO_OBJ);
-			break;
-		}
-
-		err = open_commit(&commit, pack, packidx, idx, &pid->id,
-		    objcache);
-		if (err)
-			break;
-	}
-done:
-	if (commit)
-		got_object_commit_close(commit);
-	got_object_id_queue_free(&repaint);
-
-	return err;
-}
-
 static const struct got_error *
 paint_commits(struct got_object_id_queue *ids, int *nids,
     struct got_object_idset *keep, struct got_object_idset *drop,
@@ -1860,12 +1746,6 @@ paint_commits(struct got_object_id_queue *ids, int *nids,
 				    NULL);
 				if (err)
 					goto done;
-				err = repaint_parent_commits(&qid->id, idx,
-				    COLOR_SKIP, skip, skip, ids, nids,
-				    &painted, &npainted, pack, packidx,
-				    objcache);
-				if (err)
-					goto done;
 				break;
 			}
 			if (!got_object_idset_contains(keep, &qid->id)) {
@@ -1884,12 +1764,6 @@ paint_commits(struct got_object_id_queue *ids, int *nids,
 				    NULL);
 				if (err)
 					goto done;
-				err = repaint_parent_commits(&qid->id, idx,
-				    COLOR_SKIP, skip, skip, ids, nids,
-				    &painted, &npainted, pack, packidx,
-				    objcache);
-				if (err)
-					goto done;
 				break;
 			}
 			if (!got_object_idset_contains(drop, &qid->id)) {
@@ -2024,9 +1898,8 @@ commit_painting_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	const struct got_error *err = NULL;
 	size_t datalen;
 	struct got_object_id_queue ids;
-	size_t nkeep = 0, ndrop = 0, nskip = 0;
-	int nids = 0;
-	uintptr_t color;
+	size_t nids;
+	int nqueued = 0, done = 0;
 
 	STAILQ_INIT(&ids);
 
@@ -2034,42 +1907,23 @@ commit_painting_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
 
-	color = COLOR_KEEP;
-	err = recv_object_id_queue(&nkeep, &ids, (void *)color, NULL, ibuf);
-	if (err)
-		goto done;
-	if (nids + nkeep > INT_MAX) {
-		err = got_error(GOT_ERR_PRIVSEP_LEN);
-		goto done;
+	while (!done) {
+		err = got_privsep_recv_object_id_queue(&done, &ids, &nids, ibuf);
+		if (err)
+			goto done;
+		if (nqueued + nids > INT_MAX) {
+			err = got_error(GOT_ERR_PRIVSEP_LEN);
+			goto done;
+		}
+		nqueued += nids;
 	}
-	nids += nkeep;
 
-	color = COLOR_DROP;
-	err = recv_object_id_queue(&ndrop, &ids, (void *)color, NULL, ibuf);
-	if (err)
-		goto done;
-	if (nids + ndrop > INT_MAX) {
-		err = got_error(GOT_ERR_PRIVSEP_LEN);
-		goto done;
-	}
-	nids += ndrop;
-
-	color = COLOR_SKIP;
-	err = recv_object_id_queue(&nskip, &ids, (void *)color, NULL, ibuf);
-	if (err)
-		goto done;
-	if (nids + nskip > INT_MAX) {
-		err = got_error(GOT_ERR_PRIVSEP_LEN);
-		goto done;
-	}
-	nids += nskip;
-
-	err = paint_commits(&ids, &nids, keep, drop, skip,
+	err = paint_commits(&ids, &nqueued, keep, drop, skip,
 	    pack, packidx, ibuf, objcache);
 	if (err)
 		goto done;
 
-	err = got_privsep_send_painted_commits(ibuf, &ids, &nids, 0, 1);
+	err = got_privsep_send_painted_commits(ibuf, &ids, &nqueued, 0, 1);
 	if (err)
 		goto done;
 
