@@ -485,9 +485,11 @@ test_fetch_all() {
 
 	# remove default branch from got.conf, fetch all branches
 	ed -s $testroot/repo-clone/got.conf <<-EOF
+	/fetch {/d
 	/branch {/c
 	fetch_all_branches yes
 	.
+	/}/d
 	w
 	EOF
 
@@ -2025,11 +2027,25 @@ test_fetch_basic_http() {
 	local testurl=http://127.0.0.1:$GOT_TEST_HTTP_PORT
 	local commit_id=`git_show_head $testroot/repo`
 
-	timeout 10 ./http-server -p $GOT_TEST_HTTP_PORT $testroot \
+	timeout 20 ./http-server -p $GOT_TEST_HTTP_PORT $testroot \
 	    > $testroot/http-server.log &
-	trap "kill %1" HUP INT QUIT PIPE TERM
 
-	sleep 2 # server starts up
+	sleep 1 # server starts up
+	for i in 1 2 3 4; do
+		if grep -q ': ready' $testroot/http-server.log; then
+			break
+		fi
+		if [ $i -eq 4 ]; then
+			echo "http-server startup timeout" >&2
+			test_done "$testroot" "1"
+			# timeout(1) will kill the server eventually
+			return 1
+		fi
+		sleep 1 # server is still starting up
+	done
+
+	http_pid=`head -n 1 $testroot/http-server.log | cut -d ':' -f1`
+	trap "kill -9 $http_pid; wait $http_pid" HUP INT QUIT PIPE TERM
 
 	got clone -q $testurl/repo $testroot/repo-clone
 	ret=$?
@@ -2060,8 +2076,8 @@ test_fetch_basic_http() {
 		return 1
 	fi
 
-	kill %1
-	wait %1 # wait for http-server
+	kill $http_pid
+	wait $http_pid
 
 	echo -n > $testroot/stdout.expected
 
