@@ -286,10 +286,15 @@ test_branch_created() {
 	(cd $testroot/wt && got branch newbranch > /dev/null)
 
 	echo "change alpha on branch" > $testroot/wt/alpha
-	(cd $testroot/wt && got commit -m 'newbranch' > /dev/null)
+	(cd $testroot/wt && got commit -m 'change alpha on newbranch' \
+		> /dev/null)
 	local commit_id=`git_show_branch_head $testroot/repo-clone newbranch`
 	local author_time=`git_show_author_time $testroot/repo-clone $commit_id`
 
+	echo "change alpha again" > $testroot/wt/alpha
+	(cd $testroot/wt && got commit -m 'change alpha again' > /dev/null)
+	local commit_id2=`git_show_branch_head $testroot/repo-clone newbranch`
+	local author_time2=`git_show_author_time $testroot/repo-clone $commit_id`
 	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
 		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
 
@@ -305,6 +310,7 @@ test_branch_created() {
 
 	wait %1 # wait for nc -l
 
+	short_commit_id2=`trim_obj_id 12 $commit_id2`
 	short_commit_id=`trim_obj_id 12 $commit_id`
 	HOSTNAME=`hostname`
 	printf "HELO localhost\r\n" > $testroot/stdout.expected
@@ -315,17 +321,125 @@ test_branch_created() {
 	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
 	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
 	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
-	printf "${GOTD_DEVUSER} created refs/heads/newbranch: $short_commit_id\r\n" \
+	printf "${GOTD_DEVUSER} created refs/heads/newbranch: $short_commit_id2\r\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
 	printf "commit $commit_id\n" >> $testroot/stdout.expected
 	printf "from: $GOT_AUTHOR\n" >> $testroot/stdout.expected
-	d=`date -u -r $author_time +"%a %b %e %X %Y UTC"`
+	d=`date -u -r $author_time2 +"%a %b %e %X %Y UTC"`
 	printf "date: $d\n" >> $testroot/stdout.expected
-	printf "messagelen: 11\n" >> $testroot/stdout.expected
+	printf "messagelen: 27\n" >> $testroot/stdout.expected
 	printf " \n" >> $testroot/stdout.expected
-	printf " newbranch\n \n" >> $testroot/stdout.expected
+	printf " change alpha on newbranch\n \n" >> $testroot/stdout.expected
 	printf " M  alpha  |  1+  1-\n\n"  >> $testroot/stdout.expected
+	printf "1 file changed, 1 insertion(+), 1 deletion(-)\n\n" \
+		>> $testroot/stdout.expected
+	printf "commit $commit_id2\n" >> $testroot/stdout.expected
+	printf "from: $GOT_AUTHOR\n" >> $testroot/stdout.expected
+	d=`date -u -r $author_time2 +"%a %b %e %X %Y UTC"`
+	printf "date: $d\n" >> $testroot/stdout.expected
+	printf "messagelen: 20\n" >> $testroot/stdout.expected
+	printf " \n" >> $testroot/stdout.expected
+	printf " change alpha again\n \n" >> $testroot/stdout.expected
+	printf " M  alpha  |  1+  1-\n\n"  >> $testroot/stdout.expected
+	printf "1 file changed, 1 insertion(+), 1 deletion(-)\n\n" \
+		>> $testroot/stdout.expected
+	printf "\r\n" >> $testroot/stdout.expected
+	printf ".\r\n" >> $testroot/stdout.expected
+	printf "QUIT\r\n" >> $testroot/stdout.expected
+
+	grep -v ^Date $testroot/stdout > $testroot/stdout.filtered
+	cmp -s $testroot/stdout.expected $testroot/stdout.filtered
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/stdout.expected $testroot/stdout.filtered
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	test_done "$testroot" "$ret"
+}
+
+test_branch_recreated() {
+	local testroot=`test_init branch_recreated 1`
+
+	got clone -a -q ${GOTD_TEST_REPO_URL} $testroot/repo-clone
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got clone failed unexpectedly" >&2
+		test_done "$testroot" 1
+		return 1
+	fi
+
+	got branch -r $testroot/repo-clone -d newbranch > /dev/null
+
+	got checkout -q $testroot/repo-clone $testroot/wt >/dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got checkout failed unexpectedly" >&2
+		test_done "$testroot" 1
+		return 1
+	fi
+
+	(cd $testroot/wt && got branch newbranch > /dev/null)
+
+	echo "change beta on branch" > $testroot/wt/beta
+	(cd $testroot/wt && got commit -m 'change beta on newbranch' \
+		> /dev/null)
+	local commit_id=`git_show_branch_head $testroot/repo-clone newbranch`
+	local author_time=`git_show_author_time $testroot/repo-clone $commit_id`
+
+	echo "change beta again" > $testroot/wt/beta
+	(cd $testroot/wt && got commit -m 'change beta again' > /dev/null)
+	local commit_id2=`git_show_branch_head $testroot/repo-clone newbranch`
+	local author_time2=`git_show_author_time $testroot/repo-clone $commit_id`
+	(printf "220\r\n250\r\n250\r\n250\r\n354\r\n250\r\n221\r\n" \
+		| timeout 5 nc -l "$GOTD_TEST_SMTP_PORT" > $testroot/stdout) &
+
+	sleep 1 # server starts up
+
+	got send -b newbranch -f -q -r $testroot/repo-clone
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got send failed unexpectedly" >&2
+		test_done "$testroot" "1"
+		return 1
+	fi
+
+	wait %1 # wait for nc -l
+
+	short_commit_id2=`trim_obj_id 12 $commit_id2`
+	short_commit_id=`trim_obj_id 12 $commit_id`
+	HOSTNAME=`hostname`
+	printf "HELO localhost\r\n" > $testroot/stdout.expected
+	printf "MAIL FROM:<${GOTD_USER}@${HOSTNAME}>\r\n" \
+		>> $testroot/stdout.expected
+	printf "RCPT TO:<${GOTD_DEVUSER}>\r\n" >> $testroot/stdout.expected
+	printf "DATA\r\n" >> $testroot/stdout.expected
+	printf "From: ${GOTD_USER}@${HOSTNAME}\r\n" >> $testroot/stdout.expected
+	printf "To: ${GOTD_DEVUSER}\r\n" >> $testroot/stdout.expected
+	printf "Subject: $GOTD_TEST_REPO_NAME: " >> $testroot/stdout.expected
+	printf "${GOTD_DEVUSER} changed refs/heads/newbranch: $short_commit_id2\r\n" \
+		>> $testroot/stdout.expected
+	printf "\r\n" >> $testroot/stdout.expected
+	printf "commit $commit_id\n" >> $testroot/stdout.expected
+	printf "from: $GOT_AUTHOR\n" >> $testroot/stdout.expected
+	d=`date -u -r $author_time2 +"%a %b %e %X %Y UTC"`
+	printf "date: $d\n" >> $testroot/stdout.expected
+	printf "messagelen: 26\n" >> $testroot/stdout.expected
+	printf " \n" >> $testroot/stdout.expected
+	printf " change beta on newbranch\n \n" >> $testroot/stdout.expected
+	printf " M  beta  |  1+  1-\n\n"  >> $testroot/stdout.expected
+	printf "1 file changed, 1 insertion(+), 1 deletion(-)\n\n" \
+		>> $testroot/stdout.expected
+	printf "commit $commit_id2\n" >> $testroot/stdout.expected
+	printf "from: $GOT_AUTHOR\n" >> $testroot/stdout.expected
+	d=`date -u -r $author_time2 +"%a %b %e %X %Y UTC"`
+	printf "date: $d\n" >> $testroot/stdout.expected
+	printf "messagelen: 19\n" >> $testroot/stdout.expected
+	printf " \n" >> $testroot/stdout.expected
+	printf " change beta again\n \n" >> $testroot/stdout.expected
+	printf " M  beta  |  1+  1-\n\n"  >> $testroot/stdout.expected
 	printf "1 file changed, 1 insertion(+), 1 deletion(-)\n\n" \
 		>> $testroot/stdout.expected
 	printf "\r\n" >> $testroot/stdout.expected
@@ -744,6 +858,7 @@ run_test test_file_changed
 run_test test_many_commits_not_summarized
 run_test test_many_commits_summarized
 run_test test_branch_created
+run_test test_branch_recreated
 run_test test_branch_removed
 run_test test_tag_created
 run_test test_tag_changed
